from __future__ import annotations

from typing import Dict, List, Tuple

from binarychain import BinaryChain

from .exceptions import ProtocolError
from .encoders.base import EncoderDecoderBase
from .serialisers import BaseSerialiser
from . import messages, serialisers
from .encoders import hs, v050
from .registers import FunctionRegister


class ProtocolEngine:
    def __init__(self, load_default_serialisers=True, load_default_encoders=True,
                 always_send_ids=False) -> None:
        self.serialisers: Dict[str, BaseSerialiser] = dict()
        self.encoders: Dict[str, EncoderDecoderBase] = dict()
        self.always_send_ids = always_send_ids

        if load_default_serialisers:
            self.load_default_serialisers()

        if load_default_encoders:
            self.load_default_encoders()

    def load_default_serialisers(self):
        self.add_serialiser(serialisers.JsonSerialiser())

        if serialisers.msgpack:
            self.add_serialiser(serialisers.MsgpackSerialiser())

        if serialisers.cbor2:
            self.add_serialiser(serialisers.CborSerialiser())

    def add_serialiser(self, serialiser: BaseSerialiser, sig_override: str=''):
        sig = sig_override if sig_override else serialiser.sig
        if sig in self.serialisers.keys():
            raise ValueError(f'Serialiser with sig {sig} already exists')
        if sig is None:
            raise ValueError(f'Sig for serialiser {serialiser} not set')
        self.serialisers[sig] = serialiser

    def load_default_encoders(self):
        self.load_encoder(hs.BasicEncoderDecoder(self))
        self.load_encoder(v050.V050EncoderDecoder(self))

    def load_encoder(self, engine: EncoderDecoderBase):
        self.encoders[engine.TAG] = engine

    def parse_incoming_message(self, incoming_bin_chain: BinaryChain):
        envelope_serialisation_code = incoming_bin_chain.prefix
        msg_parts = incoming_bin_chain.parts

        try:
            envelope_serialiser = self.serialisers[envelope_serialisation_code]
        except KeyError:
            raise ProtocolError(f'Serialiser {envelope_serialisation_code} not found')
        
        if not msg_parts:
            raise ProtocolError('No envelope data found - empty binary chain') 
        
        raw_envelope_data, raw_payload_data = msg_parts[0], msg_parts[1:]
        envelope_data = envelope_serialiser.deserialise(raw_envelope_data)

        # print(envelope_data)
        # print(type(envelope_data))

        if not isinstance(envelope_data, dict):
            raise ProtocolError('Envelope must be a map / dict')
        
        try:
            protocol_version = envelope_data.pop('v')
        except KeyError:
            raise ProtocolError('No protocol version not found')
        
        try:
            engine = self.encoders[protocol_version]
        except KeyError:
            raise ProtocolError(f'Engine with tag {protocol_version} not found')
        
        # decode the envelope
        message = engine.decode_raw_envelope(envelope_data, envelope_serialisation_code)

        # parse the payload data
        engine.parse_payload_data(message, raw_payload_data)

        return message        

    def message_to_binary_chain(self, message: messages.MessageBase):
        envelope_serialisation_code = message.envelope_serialisation_code
        if not envelope_serialisation_code:
            raise ValueError('Envelope Serialisation Code must be set')
        
        try:
            envelope_serialiser = self.serialisers[envelope_serialisation_code]
        except KeyError:
            raise ValueError(f'Serialiser {envelope_serialisation_code} not found')

        protocol_version = message.protocol_version
        if not protocol_version:
            raise ValueError('Protocol version must be set')
        
        try:
            engine = self.encoders[protocol_version]
        except KeyError:
            raise ProtocolError(f'Engine with tag {protocol_version} not found')

        envelope_data = engine.encode_envelope(message)
        msg_parts = [envelope_serialiser.serialise(envelope_data)]
        msg_parts.extend(engine.serialise_payload_data(message))

        return BinaryChain(envelope_serialisation_code, msg_parts)
    
    def get_system_register(self):
        register = FunctionRegister(namespace='#sys')
        register.register_func(self._ping, 'ping')
        return register

    @staticmethod
    def _ping():
        return 'pong'



from __future__ import annotations

import logging
from typing import Dict, Any, Union, Tuple, TYPE_CHECKING

from ..messages import (Acknowledge, Notification, Request, Response, ExceptionCategory, ExceptionInfo, CallBase, MessageBase, ResponseBase,
                        RequestCallbackInfo, RequestType, ResponseType, ExceptionResponse, IterableCallbackInfo, FinalType)

from .base import EncoderDecoderBase, make_out_except_data
from ..exceptions import ProtocolError

if TYPE_CHECKING:
    from ..engine import ProtocolEngine


log = logging.getLogger(__name__)


class V050EncoderDecoder(EncoderDecoderBase):
    TAG = '.5'
    max_bc_length = 3   # envelope, params or result, callback params (prefix not counted)
    envelope_marker_map = {'r': (Request, 'target', set('qs')),
                           'n': (Notification, 'target', None),
                           'p': (Response, 'request_id', set('m')),
                           'x': (ExceptionResponse, 'request_id', None),
                           'a': (Acknowledge, 'request_id', None)
                           }
    envelope_unique = set(envelope_marker_map.keys())
    envelope_sup = {'i': 'message_id',  # always allowed, but may be ignored
                    's': 'namespace',
                    'd': 'node',
                    'c': 'callback_request_id',
                    'v': 'protocol_version',  # required
                    'l': 'payload_serialisation_code',   # if missing, same as envelope serialisation code
                    }
    flags = {'a': 'acknowledge'}
    payload_info_key = 'f'  # if missing, default payload for message type
                            # FIXME: Not used yet. Use for binary in requests?

    def __init__(self, protocol_engine: ProtocolEngine):
        EncoderDecoderBase.__init__(self, protocol_engine)
        self._env_make_map = {Request: self._make_request_env,
                              Notification: self._make_notification_env,
                              Response: self._make_response_env,
                              ExceptionResponse: self._make_exception_env,
                              Acknowledge: self._make_ack_env,
                              }
        
        self._payload_make_map = {Request: self._make_request_payload,
                                  Notification: self._make_notification_payload,
                                  Response: self._make_response_payload,
                                  ExceptionResponse: self._make_exception_payload,
                                  Acknowledge: self._make_ack_payload,
                                  }
    
    def _get_payload_serialiser(self, message: MessageBase):
        # get the serialisers
        psc = message.envelope_serialisation_code
        if isinstance(message, (CallBase, ResponseBase)):
            if message.payload_serialisation_code: 
                psc = message.payload_serialisation_code

        if not psc:
            raise ValueError('no serialisation code set')

        try:
            payload_serialiser = self.protocol_engine.serialisers[psc]
        except KeyError:
            raise ValueError(f'Serialiser "{psc}" not found')

        return payload_serialiser
        
    def serialise_payload_data(self, message: MessageBase):
        try:
            # noinspection PyTypeChecker
            make_method = self._payload_make_map[type(message)]
            return make_method(message)
        except KeyError:
            raise TypeError(f'Invalid message type {type(message)}')
    
    def encode_envelope(self, message: MessageBase):
        try:
            # noinspection PyTypeChecker
            make_method = self._env_make_map[type(message)]
            return make_method(message)
        except KeyError:
            raise TypeError(f'Invalid message type {type(message)}')

    def _make_request_env(self, message: Request):
        if message.message_id is None:
            raise ValueError('message id is required on Outgoing Requests')
        marker_chrs = ['r']
        if message.request_type == RequestType.SILENT and message.acknowledge:
            raise ValueError('acknowledge can not use set on SILENT requests.')
        if message.acknowledge:
            marker_chrs.append('a')
        if message.request_type != RequestType.NORMAL:
            marker_chrs.append(str(message.request_type))
        envelope_data = {''.join(marker_chrs): message.target, 'i': message.message_id, 'v': message.protocol_version}
        if message.callback_request_id:
            envelope_data['c'] = message.callback_request_id
        if message.node:
            envelope_data['d'] = message.node
        if message.namespace:
            envelope_data['s'] = message.namespace
        if message.payload_serialisation_code != message.envelope_serialisation_code:
            envelope_data['l'] = message.payload_serialisation_code
        return envelope_data

    def _make_request_payload(self, message: Request):
        payload_parts = list()
        payload_serialiser = self._get_payload_serialiser(message)

        if message.params or message.callback_params:
            param_part = payload_serialiser.serialise(message.params) if message.params else b''
            payload_parts.append(param_part)

            if message.callback_params:
                cb_dict = {}
                icb_dict = {}
                for param_name, cb_info in message.callback_params.items():
                    if isinstance(cb_info, RequestCallbackInfo):
                        cb_dict[param_name] = dict(p=cb_info.param_details)
                    elif isinstance(cb_info, IterableCallbackInfo):
                        icb_dict[param_name] = dict(r=cb_info.return_details)
                    else:
                        raise TypeError(f'Invalid callback param type {type(cb_info)}')
                special_params = {'#cb': cb_dict, '#icb': icb_dict}
                special_params_part = payload_serialiser.serialise(special_params)
                payload_parts.append(special_params_part)
        return payload_parts
        
    def _make_notification_env(self, message: Notification):
        envelope_data: Dict[str, Any] = {'n': message.target, 'v': message.protocol_version}
        if message.node:
            envelope_data['d'] = message.node
        if message.namespace:
            envelope_data['s'] = message.namespace
        if message.callback_request_id:
            envelope_data['c'] = message.callback_request_id
        if message.payload_serialisation_code != message.envelope_serialisation_code:
            envelope_data['l'] = message.payload_serialisation_code
        return envelope_data
    
    def _make_notification_payload(self, message: Notification):
        payload_parts = list()
        payload_serialiser = self._get_payload_serialiser(message)

        if message.params:
            param_part = payload_serialiser.serialise(message.params)
            payload_parts.append(param_part)
        return payload_parts

    def _make_response_env(self, message: Response):
        marker = 'p'
        marker += message.response_type
        if message.final:
            marker += message.final
        if message.acknowledge:
            if not message.message_id:
                raise ValueError('A message id must be given when requesting an ack')
            marker += 'a'
        envelope_data: Dict[str, Any] = {marker: message.request_id, 'v': message.protocol_version}
        if message.message_id is not None:
            envelope_data['i'] = message.message_id
        if message.payload_serialisation_code != message.envelope_serialisation_code:
            envelope_data['l'] = message.payload_serialisation_code
        return envelope_data
    
    def _make_response_payload(self, message: Response):
        payload_parts = list()
        payload_serialiser = self._get_payload_serialiser(message)

        if message.result is not None:  # FIXME? Do we need to allow returning of None, and use Void instead?
            payload_parts.append(payload_serialiser.serialise(message.result))
        return payload_parts

    def _make_exception_env(self, message: ExceptionResponse):
        marker = 'x'
        if message.acknowledge:
            if not message.message_id:
                raise ValueError('A message id must be given when requesting an ack')
            marker += 'a'
        envelope_data: Dict[str, Any] = {marker: message.request_id, 'v': message.protocol_version}
        if message.message_id is not None:
            envelope_data['i'] = message.message_id
        if message.payload_serialisation_code != message.envelope_serialisation_code:
            envelope_data['l'] = message.payload_serialisation_code
        return envelope_data

    def _make_exception_payload(self, message: ExceptionResponse):
        except_data = make_out_except_data(message)
        payload_serialiser = self._get_payload_serialiser(message)
        return [payload_serialiser.serialise(except_data)]
    
    def _make_ack_env(self, message: Acknowledge):
        return {'a': message.request_id, 'v': message.protocol_version}
    
    def _make_ack_payload(self, message: Acknowledge):
        return []

    def parse_payload_data(self, message: Union[CallBase, ResponseBase], raw_payload_data: list):
        assert isinstance(message, (CallBase, ResponseBase))
        payload_serialisation_code = message.payload_serialisation_code if message.payload_serialisation_code else message.envelope_serialisation_code
        if not payload_serialisation_code:
            raise ProtocolError('No envelope serialisation code set')
        
        try:
            serialiser = self.protocol_engine.serialisers[payload_serialisation_code]
        except KeyError:
            raise ProtocolError(f'Serialiser for {payload_serialisation_code} not found')

        if isinstance(message, Request):
            if not raw_payload_data:
                return message

            # normal params
            if raw_payload_data[0]:
                message.params = serialiser.deserialise(raw_payload_data[0])
            if len(raw_payload_data) > 1:
                special_params_dict = serialiser.deserialise(raw_payload_data[1])  # cb_param -> target_name
                cb_params = dict()
                raw_std_cb_params = special_params_dict.get('#cb', dict())
                for (param_name, cb_info) in raw_std_cb_params.items():   # FIXME: currently ignoring info
                    cb_params[param_name] = {'#cb': {'callback_request_id': message.message_id, 'cb_info': cb_info}}
                raw_iter_cb_params = special_params_dict.get('#icb', dict())
                for (param_name, cb_info) in raw_iter_cb_params.items():
                    cb_params[param_name] = {'#icb': {'callback_request_id': message.message_id, 'cb_info': cb_info}}
                message.callback_params = cb_params
            if len(raw_payload_data) > 2:
                raise ProtocolError('Extra payload data found')
            return message
        
        if isinstance(message, Notification):            
            message.params = serialiser.deserialise(raw_payload_data[0])
            return message
        
        if isinstance(message, Response):
            if raw_payload_data:
                message.result = serialiser.deserialise(raw_payload_data[0])
            return message
        
        if isinstance(message, ExceptionResponse):
            raw_except_data = serialiser.deserialise(raw_payload_data[0])
            message.exc_info = ExceptionInfo(ExceptionCategory(raw_except_data['cat']), raw_except_data['type'], raw_except_data['msg'])
            return message

        if isinstance(message, Acknowledge):
            return message
        
        raise RuntimeError('invalid message passed in')

    def decode_raw_envelope(self, envelope_data: dict, envelope_serialisation_code: str):
        if not isinstance(envelope_data, dict):
            raise ProtocolError('envelope must be a dict')
        key_type_to_flags = {key[0]: key[1:] for key in envelope_data.keys()}
        key_type_set = set(key_type_to_flags.keys())
        marker_list = list(key_type_set.intersection(self.envelope_unique))
        if len(marker_list) != 1:
            log.error(f'marker_list is {marker_list} but should be a single item')
            raise ProtocolError('single marker not found')
        marker = marker_list[0]
        cls, attr, valid_type_flags = self.envelope_marker_map[marker]
        flags = key_type_to_flags[marker]
        type_flags = set()
        if valid_type_flags:
            assert isinstance(valid_type_flags, set)
            type_flags = valid_type_flags.intersection(set(flags))
            if len(type_flags) > 1:
                raise ProtocolError('more than one type-flag found')
        actual_key = marker + flags
        kwargs = {attr: envelope_data.pop(actual_key)}
        for key, attr in self.envelope_sup.items():
            if key in envelope_data:
                kwargs[attr] = envelope_data.pop(key)
        if type_flags:
            attr, value = self._get_request_response_type(cls, type_flags)
            kwargs[attr] = value
        
        # set the envelope_serialisation_code as it is not in the envelope data itself 
        kwargs['envelope_serialisation_code'] = envelope_serialisation_code
        kwargs['protocol_version'] = self.TAG

        result = cls(**kwargs)
        for flag in flags:
            if flag == 'a':
                if hasattr(result, 'acknowledge'):
                    result.acknowledge = True
                else:
                    log.warning('Got unexpected acknowledge flag')  # FIXME: make these warning protocol errors?
            elif flag in ('f', 't'):
                if hasattr(result, 'final'):
                    result.final = FinalType(flag)
                else:
                    log.warning('Got unexpected final flag')
        if envelope_data:
            raise ProtocolError(f'extra keys found: {envelope_data.keys()}')
        return result
    
    def _get_request_response_type(self, cls: type, type_flags: set) -> Tuple[str, Union[RequestType, ResponseType]]:
        if len(type_flags) != 1:
            raise ValueError('Invalid type_flags state')
        type_chr = type_flags.pop()
        if cls is Request:
            return 'request_type', RequestType(type_chr)
        if cls is Response:
            return 'response_type', ResponseType(type_chr)
        raise ValueError(f'Invalid cls of {cls!s}')

    # def get_system_register(self):
    #     register = FunctionRegister(namespace='#sys')
    #     register.register_func(self._ping, 'ping')
    #     register.register_func(self.get_engine_signature, 'get_sig')
    #     return register
    
    # def get_engine_signature(self):
    #     return self.TAG
    
    # def _ping(self):
    #     return 'pong'



from __future__ import annotations

from typing import Any, Iterable, Optional, Callable, Dict
import sys

if sys.version_info >= (3, 11) or sys.implementation.name == 'micropython':
    from enum import StrEnum
else:
    from backports.strenum import StrEnum # type: ignore


__all__ = ['RequestType', 'ResponseType', 'FinalType', 'PushIterableInfo', 'MessageBase', 'CallBase', 'Request',
           'Notification', 'ResponseBase', 'Response', 'MessageSentEvent', 'Acknowledge',
           'ExceptionCategory', 'ExceptionInfo', 'ExceptionResponse', 'RequestCallbackInfo', 'IterableCallbackInfo']


class RequestType(StrEnum):
    NORMAL = ''
    QUIET = 'q'    # returns an empty response after the call is completed, unless there is an exception.
                   # (Like normal, but flagging that we don't want the result.)
    SILENT = 's'   # returns an empty response after call is *delivered*,
                   # unless there is an exception (transport error / invalid target).


class ResponseType(StrEnum):
    NORMAL = ''      # single response only
    MULTIPART = 'm'  # may return multiple response messages. The last one is flagged as 'final'
                     # - either FINAL (has data) or TERMINATOR (no data)


class FinalType(StrEnum):
    FINAL = 'f'       # final - with data
    TERMINATOR = 't'  # terminator - no data


# May no longer need this?

# # Python makes no distinction between a function that has no return value,
# # and a function that returns None.
# # Add this if it makes it more consistent with other languages.

# class VoidType:
#     def __repr__(self):
#         return 'Void'


# Void = VoidType()


class PushIterableInfo:
    def __init__(self, iter: Optional[Iterable] = None, return_details: Any = None):
        self.iter = iter
        self.return_details = return_details     # FIXME: return_details is not done yet.


class EventBase:
    _prop_cache = None

    @property
    def properties(self):
        if self._prop_cache is None:
            # FIXME?: Is this safe under free threading?
            self._prop_cache = sorted([key for key in self.__dict__.keys() if not key.startswith('_')])
        return self._prop_cache
    
    def __repr__(self):
        items = [f"{prop}={getattr(self, prop)!r}" for prop in self.properties]
        if items:
            return f"<{self.__class__.__name__}: {', '.join(items)}>"
        else:
            return f"<{self.__class__.__name__}>"
        
    def as_dict(self):
        return {(prop, getattr(self, prop)) for prop in self.properties}
        
    def __eq__(self, other: EventBase):
        if type(self) == type(other):
            return self.as_dict() == other.as_dict()
        else:
            return False


class TransportEvent(EventBase):
    pass


class MessageSentEvent(TransportEvent):
    def __init__(self, request_id: int):
        self.request_id = request_id


class TransportClosedEvent(TransportEvent):
    def __init__(self, closing_end: str):
        # FIXME: Use a StrEnum?
        if closing_end not in ('remote', 'local'):
            raise ValueError()
        self.closing_end = closing_end
        
    def raise_as_exception(self):
        raise TransportClosedError(self.closing_end)


class TransportBrokenEvent(TransportEvent):
    def __init__(self, original_exc: Exception):
        self.original_exc = original_exc

    def raise_as_exception(self):
        raise TransportBrokenError(self.original_exc)


class TransportClosedError(TransportClosedEvent, Exception):
    pass


class TransportBrokenError(TransportBrokenEvent, Exception):
    pass


class MessageBase(EventBase):
    def __init__(self, envelope_serialisation_code: Optional[str], protocol_version: Optional[str]):
        self.envelope_serialisation_code = envelope_serialisation_code
        self.protocol_version = protocol_version


class CallBase(MessageBase):
    def __init__(self, target: str, namespace: str='', node: Optional[str]=None,
                 params: Optional[Dict]=None,  # param name -> value
                 callback_request_id: Optional[int]=None, envelope_serialisation_code: Optional[str]=None,
                 protocol_version: Optional[str]=None, payload_serialisation_code: Optional[str]=None,
                 ):
        MessageBase.__init__(self, envelope_serialisation_code, protocol_version)
        self.target = target
        self.namespace = '' if not namespace else namespace  # FIXME: Review if None is better as the default
        self.node = node
        self.params = params
        self.payload_serialisation_code = payload_serialisation_code
        self.callback_request_id = callback_request_id


class Request(CallBase):
    def __init__(self, target: str, namespace: str='', node: Optional[str]=None,
                 params: Optional[dict]=None,  # param name -> value
                 callback_request_id: Optional[int]=None, envelope_serialisation_code: Optional[str]=None,
                 protocol_version: Optional[str]=None,
                 payload_serialisation_code: Optional[str]=None, 
                 acknowledge: bool=False, request_type: RequestType=RequestType.NORMAL, 
                 callback_params: Optional[Dict[str, RequestCallbackInfo]]=None,  # param name -> CallbackInfo
                 message_id: Optional[int]=None
                 ):
        CallBase.__init__(self, target, namespace, node, params, callback_request_id, envelope_serialisation_code,
                          protocol_version, payload_serialisation_code)
        self.acknowledge = acknowledge
        self.request_type = request_type
        self.callback_params = callback_params
        self.message_id = message_id

        normal_params = set(self.params.keys()) if self.params else set()
        cb_params = set(self.callback_params.keys()) if self.callback_params else set()
        duplicates = normal_params.intersection(cb_params)
        if duplicates:
            raise ValueError(f'duplicate params found: {duplicates}')

        # FIXME: Do we will need this restriction now we are event based?
        if self.request_type == RequestType.SILENT and cb_params:
            raise ValueError("Can't have callbacks on Slient Requests")


# Notifications are requests that should never return a response (even in the case of errors at the remote end)
class Notification(CallBase):
    pass


class ResponseBase(MessageBase):
    def __init__(self, request_id: int, envelope_serialisation_code: Optional[str]=None, protocol_version: Optional[str]=None, 
                 payload_serialisation_code: Optional[str]=None):
        MessageBase.__init__(self, envelope_serialisation_code, protocol_version)
        self.request_id = request_id
        self.payload_serialisation_code = payload_serialisation_code


class Acknowledge(ResponseBase):
    pass

    

class Response(ResponseBase):
    def __init__(self, request_id: int, envelope_serialisation_code: str, protocol_version: str,
                 result: Any=None, 
                 payload_serialisation_code: Optional[str]=None,
                 response_type: ResponseType=ResponseType.NORMAL, acknowledge: bool=False,
                 final: Optional[FinalType]=None,   # only relevant to multipart responses
                 message_id: Optional[int]=None     # only needed if acknowledge is set
                 ):
        ResponseBase.__init__(self, request_id, envelope_serialisation_code, protocol_version,
                              payload_serialisation_code=payload_serialisation_code)
        self.result = result
        self.response_type = response_type
        self.acknowledge = acknowledge
        self.final = final
        self.message_id = message_id


class ExceptionCategory(StrEnum):
    CALLER = 'r'  # the caller made an error (invalid param value or type, not authorised, target not found)
    CALLEE = 'e'  # a callee error occurred
    TRANSPORT = 't'  # some kind of transport error, or parse error.  # FIXME: Do we need this? Does this make sense?


class ExceptionInfo:
    def __init__(self, category: ExceptionCategory, exc_type: str, msg: str, details: str = '',
                 value: Optional[int]=None):
        self.category = category
        self.exc_type = exc_type
        self.msg = msg
        self.details = details
        self.value = value


class ExceptionResponse(ResponseBase):
    def __init__(self, request_id: int, exc_info: ExceptionInfo=None, acknowledge: bool=False,
                  envelope_serialisation_code: Optional[str]=None, protocol_version: Optional[str]=None,
                 payload_serialisation_code: Optional[str] = None,
                 message_id: Optional[int]=None     # only needed if acknowledge is set
                  ):
        ResponseBase.__init__(self, request_id, envelope_serialisation_code, protocol_version,
                              payload_serialisation_code=payload_serialisation_code)
        self.exc_info = exc_info
        self.acknowledge = acknowledge
        self.message_id = message_id


# # FIXME: Should we even have these? How can they work in an event based system? Maybe for start-tls etc?
# class ControlMsg:
#     def __init__(self, msg: str='', data: bytes = b''):
#         self.msg = msg
#         self.data = data
#         self.msg.encode('ascii')  # raises an exception of not a valid ControlMsg (ie, non ascii)


class RequestCallbackInfo:
    def __init__(self, func: Callable, param_details: Optional[list[str]]=None):
        self.func = func
        self.param_details = param_details    # None = unspecified.
                                              # FIXME: just using None for now. Do we want to send types too?


class IterableCallbackInfo:
    def __init__(self, iter: Iterable, return_details: Any=None):
        self.iter = iter
        self.return_details = return_details   # FIXME: This is not done yet.



from __future__ import annotations

from abc import ABC, abstractmethod
from collections import defaultdict, namedtuple
import logging
import os
import tempfile
import sys

if 'threaded' in __name__: #= remove
    from threading import Thread, Lock, Event, current_thread  #= threaded <
    from queue import Queue, Empty as QueueEmpty  #= threaded <
else: #= remove
    import asyncio  #= async <
    from asyncio import Lock, Event  #= async <
    try: #= async <
        from asyncio import Queue, QueueEmpty #= async <
    except (AttributeError, ImportError):  #= async <
        from uasync.queues import Queue, QueueEmpty # type: ignore (for micropython)  #= async <

if sys.platform != 'webassembly':
    import socket
else:
    socket = None

running_micropython = (sys.implementation.name == 'micropython')

if not running_micropython:
    import signal


from binarychain import BinaryChain, ChainReader

from ...bases import DEFAULT_SERIALISER_SIG, DEFAULT_ENCODER_TAG
from ...engine import ProtocolEngine
from ...encoders import hs
if not running_micropython:
    from ...lib import SignalHandlerInstaller, detect_unix_socket_in_use
from ...messages import Request, Response, TransportClosedEvent, TransportBrokenEvent

from ..._template_ import _Template_MsgChannel, _Template_TransportBase


log = logging.getLogger(__name__)



#= threaded start
class _Template_TcpTransport(_Template_TransportBase):
    def __init__(self, conn_socket: socket.socket, max_msg_size=10*1024*1024, max_bc_length=10, incoming_msg_queue_size=10, outgoing_msg_queue_size=10, socket_buf_size=8192):
        super().__init__(max_msg_size, max_bc_length, incoming_msg_queue_size, outgoing_msg_queue_size, socket_buf_size)
        self.socket = conn_socket
        self.closing_socket = False

    def _read_data(self, size):
        try:
            data = self.socket.recv(size)
            if not data:
                return TransportClosedEvent('remote')
            else:
                return data
        except Exception as e:
            if self.closing_socket:
                # windows and micropython return errors once the socket is closed
                # just return 
                return TransportClosedEvent('local')
            else:
                return TransportBrokenEvent(e)

    def _write_data(self, data):
        log.debug(f'Sending data: {data}')
        self.socket.sendall(data)

    def close(self):
        self.closing_socket = True
        self.socket.shutdown(socket.SHUT_RDWR)
        self.socket.close()
#= threaded end

#= async start
class AsyncTcpTransport(_Template_TransportBase):
    def __init__(self, stream_reader: asyncio.StreamReader, stream_writer: asyncio.StreamWriter, max_msg_size=10*1024*1024, max_bc_length=10, incoming_msg_queue_size=10, outgoing_msg_queue_size=10, socket_buf_size=8192):
        super().__init__(max_msg_size, max_bc_length, incoming_msg_queue_size, outgoing_msg_queue_size, socket_buf_size)
        self.stream_reader = stream_reader
        self.stream_writer = stream_writer
        self.closing_socket = False

    async def _read_data(self, size):
        try:
            data = await self.stream_reader.read(size)
            if not data:
                return TransportClosedEvent('remote')
            else:
                return data
        except Exception as e:
            if self.closing_socket:
                # windows and micropython return errors once the socket is closed
                # just return 
                return TransportClosedEvent('local')
            else:
                return TransportBrokenEvent(e)

    async def _write_data(self, data):
        log.debug(f'Sending data: {data}')
        self.stream_writer.write(data)
        await self.stream_writer.drain()

    async def close(self):
        self.closing_socket = True
        self.stream_writer.close()
        await self.stream_writer.wait_closed()
#= async end
    
class _Template_TcpConnector:
    def __init__(self, engine: ProtocolEngine, dispatcher, default_serialiser_sig: str=DEFAULT_SERIALISER_SIG,
                 default_encoder_tag: str=DEFAULT_ENCODER_TAG, func_registers=None):
        self.engine = engine
        self.dispatcher = dispatcher
        self.default_serialiser_sig = default_serialiser_sig
        self.default_encoder_tag = default_encoder_tag
        self.func_registers = func_registers
        self.initiator = True
        self.unix_socket_path = None
        
    async def connect(self, host, port):
        if port is None:
            self.unix_socket_path = host
        #= threaded start
        if self.unix_socket_path:
            conn_socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            conn_socket.connect(self.unix_socket_path)
        else:
            conn_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            remote_address = (host, port)
            conn_socket.connect(remote_address)
        transport = _Template_TcpTransport(conn_socket)
        #= threaded end
        #= async start
        if self.unix_socket_path:
            reader, writer = await asyncio.open_unix_connection(path=self.unix_socket_path)
        else:
            reader, writer = await asyncio.open_connection(host, port)
        transport = _Template_TcpTransport(reader, writer)
        #= async end
        await transport.start()
        return _Template_MsgChannel(transport, default_serialiser_sig=self.default_serialiser_sig,
                                    default_encoder_tag=self.default_encoder_tag, initiator=self.initiator,
                                    engine=self.engine, dispatcher=self.dispatcher, func_registers=self.func_registers)


class _Template_UnixSocketConnector(_Template_TcpConnector):
    def __init__(self, engine_choices, dispatcher, default_serialiser_sig: str=DEFAULT_SERIALISER_SIG,
                 default_encoder_tag: str=DEFAULT_ENCODER_TAG, func_registers=None, handshake_cls=None):
        super().__init__(engine_choices, dispatcher, default_serialiser_sig, default_encoder_tag, func_registers)

    async def connect(self, path):
        return await super().connect(path, None)


class _Template_TcpListener:
    def __init__(self, engine: ProtocolEngine, dispatcher, func_registers=None):
        self.engine = engine
        self.dispatcher = dispatcher
        self.func_registers = func_registers
        self.initiator = False
        self.tcp_server = None
        self.address = None
        self.channel_threads = dict()  # remote_address -> thread  #= threaded
        self.channel_tasks = dict()  # remote_address -> task  #= async
        self.connected_channels = dict()  # remote_address -> channel
        self.listening_thread = None  #= threaded
        self.listening_task = None  #= async
        self.time_to_stop = Event()
        self.unix_socket_path = None
        self.replace_unix_socket_if_in_use = False
        self.server_socket = None
        self.async_server = None

    async def _detect_unix_socket_in_use(self, socket_path):
        #= threaded start
        detecting_socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        try:
            detecting_socket.connect(socket_path)
            detecting_socket.close()
            return True
        except Exception:
            return False
        #= threaded end
        #= async start
        try:
            reader, writer = await asyncio.open_unix_connection(path=self.unix_socket_path)
            writer.close()
            await writer.wait_closed()
            return True
        except Exception:
            return False
        #= async end

    async def run_server(self, bind_address, port, block=True):
        #= threaded start
        self.listening_thread = Thread(target=self._run_server, args=(bind_address, port))
        self.listening_thread.start()
        #= threaded end
        #= async start
        self.listening_task = asyncio.create_task(self._run_server(bind_address, port))
        #= async end
        if block:
            await self.block()

    async def _run_server(self, bind_address, port):
        if port is None:
            self.address = bind_address
            self.unix_socket_path = bind_address
        else:
            self.address = f'{bind_address}:{port}'
        #= threaded start
        if self.unix_socket_path:
            if os.path.exists(self.unix_socket_path) and not self.replace_unix_socket_if_in_use:
                if detect_unix_socket_in_use(self.unix_socket_path):
                    raise OSError(f'Unix Socket {self.unix_socket_path} in use')
                
            # get a temporary name in the same directory as the final path
            socket_dir = os.path.dirname(self.unix_socket_path)
            with tempfile.NamedTemporaryFile(dir=socket_dir, delete=True) as f:
                temp_uds = f.name
            self.server_socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            self.server_socket.bind(temp_uds)

            # renames are atomic on unix / linux
            os.rename(temp_uds, self.unix_socket_path)
        else:
            self.server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            address_and_port = (bind_address, port)
            self.server_socket.bind(address_and_port)
        self.server_socket.listen(5)
        self.server_socket.settimeout(1)
        log.info(f'Listing on {self.address}')
        while not self.time_to_stop.is_set():
            try:
                client_socket, remote_address = self.server_socket.accept()
            except socket.timeout:
                continue

            log.info(f'Accepted connection from {remote_address}')
            transport = _Template_TcpTransport(client_socket)
            channel_thread = Thread(target=self._start_channel, args=(transport, remote_address))
            self.channel_threads[remote_address] = channel_thread
            channel_thread.start()
        #= threaded end
        #= async start
        if sys.implementation.name == 'micropython':
            if self.unix_socket_path:
                raise RuntimeError('Unix sockets not supported on micropython')
            self.async_server = await asyncio.start_server(self._async_client_connected, bind_address, port)
            await self.async_server.wait_closed()
        else:
            if self.unix_socket_path:
                self.async_server = await asyncio.start_unix_server(self._async_client_connected, path=self.unix_socket_path)
            else:
                self.async_server = await asyncio.start_server(self._async_client_connected, bind_address, port, reuse_address=True)
            log.info(f'Listing on {self.address}')
            await self.async_server.wait_closed()
        #= async end

    #= async start
    async def _async_client_connected(self, async_reader: asyncio.StreamReader, async_writer: asyncio.StreamWriter):
        # FIXME: work out what the peername format is in micropython
        remote_address = repr(async_writer.get_extra_info('peername'))
        log.info(f'REMOTE ADDRESS: {remote_address}')
        transport = _Template_TcpTransport(async_reader, async_writer)
        channel_task = asyncio.create_task(self._start_channel(transport, remote_address))
        self.channel_tasks[remote_address] = channel_task

    #= async end
    async def shutdown_server(self):
        log.info('Starting Server Shutdown')
        #= threaded start
        self.time_to_stop.set()
        self.listening_thread.join()
        #= threaded end
        #= async start
        self.listening_task.cancel()
        try:
            await self.listening_task
        except asyncio.CancelledError:
            pass
        #= async end
        for channel in self.connected_channels.values():
            assert isinstance(channel, _Template_MsgChannel)
            await channel.shutdown_channel()
        #= threaded start
        for channel_thread in self.channel_threads.values():
            assert isinstance(channel_thread, Thread)
            channel_thread.join()
        self.server_socket.close()
        #= threaded end
        # cleanup unix socket if we are not replacing by default
        if self.unix_socket_path and not self.replace_unix_socket_if_in_use:
            os.remove(self.unix_socket_path)
        log.info('Server Shutdown Complete')

    async def _start_channel(self, transport: _Template_TcpTransport, remote_address):
        await transport.start()
        channel = _Template_MsgChannel(transport, initiator=False, engine=self.engine, dispatcher=self.dispatcher, func_registers=self.func_registers)
        self.connected_channels[remote_address] = channel
        await channel.start_channel()

    def _signal_handler(self, signum, frame):
        signame = signal.Signals(signum).name
        log.info(f'Signal handler called with signal {signame} ({signum})')
        self.time_to_stop.set()  #= threaded
        #= async start
        self.listening_task.cancel()
        #= async end

    async def block(self, signals=None):
        signal_handler_installer = None
        if not running_micropython:
            signal_handler_installer = SignalHandlerInstaller(signals)
            signal_handler_installer.install(self._signal_handler)
        try:
            #= threaded start
            while True:
                # the timeout is required for Windows to enable the signals to be caught
                if self.time_to_stop.wait(timeout=0.5):
                    break
            self.listening_thread.join()
            #= threaded end
            #= async start
            try:
                await self.listening_task
            except asyncio.CancelledError:
                pass
            #= async end
        finally:
            if signal_handler_installer:
                signal_handler_installer.remove()
        await self.shutdown_server()


class _Template_UnixSocketListener(_Template_TcpListener):
    def __init__(self, engine_choices, dispatcher, func_registers=None):
        super().__init__(engine_choices, dispatcher, func_registers)

    async def run_server(self, path, replace_if_in_use=False, block=True):
        self.replace_unix_socket_if_in_use = replace_if_in_use
        await super().run_server(path, None, block)


class StreamTransport(_Template_TransportBase):
    def __init__(self, incoming_stream, outgoing_stream, max_msg_size=10*1024*1024, max_bc_length=10, incoming_msg_queue_size=10, outgoing_msg_queue_size=10, read_buf_size=8192):
        super().__init__(max_msg_size, max_bc_length, incoming_msg_queue_size, outgoing_msg_queue_size, read_buf_size)
        self.incoming_stream = incoming_stream
        self.outgoing_stream = outgoing_stream

    async def _read_data(self, size):
        try:
            return self.incoming_stream.read1(size)  # use read1 so we don't block waiting for 'more' data
        except Exception as e:
            log.debug(f'Error reading from incoming stream in PID {os.getpid()}: {str(e)}')
            return None
    
    async def _write_data(self, data):
        # log.debug(f'Wrote to outgoing stream: {data!s}')
        self.outgoing_stream.write(data)
        self.outgoing_stream.flush()

    async def close(self):
        self.incoming_stream.close()
        self.outgoing_stream.close()


class SubprocessRunnerBase(ABC):
    def __init__(self, engine, dispatcher, func_registers=None):
        assert isinstance(engine, ProtocolEngine)
        self.engine = engine
        self.dispatcher = dispatcher
        self.func_registers = func_registers

    @abstractmethod
    async def get_channel(self, child_stdin=None, child_stdout=None):
        raise NotImplementedError()    


class ParentSubprocessRunner(SubprocessRunnerBase):
    async def get_channel(self, child_stdin, child_stdout):
        transport = StreamTransport(outgoing_stream=child_stdin, incoming_stream=child_stdout)
        return _Template_MsgChannel(transport, initiator=True, engine=self.engine, dispatcher=self.dispatcher, func_registers=self.func_registers)


class ChildSubprocessRunner(SubprocessRunnerBase):
    def get_channel(self):
        transport = StreamTransport(sys.stdin.buffer, sys.stdout.buffer)
        return _Template_MsgChannel(transport, initiator=False, engine=self.engine, dispatcher=self.dispatcher, func_registers=self.func_registers)



# -----------------------------------------------------------------------
# WARNING: This file is auto-generated and should not be edited manually!
#
# It is generated using 'generate_from_template_code.py' using the
# _template_ directory as the source.
# -----------------------------------------------------------------------


import logging
import signal

from threading import Thread
from websockets.sync import client
from websockets.sync import server

import websockets.exceptions

from ...engine import ProtocolEngine
from ...lib import SignalHandlerInstaller
from ...threaded import ThreadedMsgChannel
from ..transports import ThreadedTransportBase
from ...bases import DEFAULT_SERIALISER_SIG, DEFAULT_ENCODER_TAG

log = logging.getLogger(__name__)


class ThreadedWebsocketTransport(ThreadedTransportBase):
    def __init__(self, websocket, max_msg_size=10*1024*1024, max_bc_length=10, incoming_msg_queue_size=10,
                 outgoing_msg_queue_size=10, socket_buf_size=8192):
        super().__init__(max_msg_size, max_bc_length, incoming_msg_queue_size, outgoing_msg_queue_size, socket_buf_size)
        self.websocket = websocket

    def _read_data(self, size):
        try:
            return self.websocket.recv()
        except websockets.exceptions.ConnectionClosed:
            return ''

    def _write_data(self, data):
        log.debug(f'Sending data: {data}')
        self.websocket.send(data)

    def close(self):
        self.websocket.close()

    
class ThreadedWebsocketConnector:
    def __init__(self, engine, dispatcher, default_serialiser_sig: str=DEFAULT_SERIALISER_SIG,
                 default_encoder_tag: str=DEFAULT_ENCODER_TAG, func_registers=None, handshake_cls=None):
        self.engine = engine
        self.dispatcher = dispatcher
        self.default_serialiser_sig = default_serialiser_sig
        self.default_encoder_tag = default_encoder_tag
        self.func_registers = func_registers
        self.initiator = True

    def connect(self, ws_uri):
        websocket = client.connect(ws_uri)
        transport = ThreadedWebsocketTransport(websocket)  # FIX_ME: Allow options to be set / passed in??
        return ThreadedMsgChannel(transport, default_serialiser_sig=self.default_serialiser_sig,
                                    default_encoder_tag=self.default_encoder_tag, initiator=self.initiator,
                                    engine=self.engine, dispatcher=self.dispatcher, func_registers=self.func_registers)


class ThreadedWebsocketListener:
    def __init__(self, engine: ProtocolEngine, dispatcher, default_serialiser_sig: str=DEFAULT_SERIALISER_SIG,
                 default_encoder_tag: str=DEFAULT_ENCODER_TAG, func_registers=None):
        self.engine = engine
        self.dispatcher = dispatcher
        self.default_serialiser_sig = default_serialiser_sig
        self.default_encoder_tag = default_encoder_tag
        self.func_registers = func_registers
        self.initiator = False
        self.websocket_server = None
        self.listening_thread = None
        self.connected_channels = dict()  # remote_address -> channel
        self.time_to_stop = False

    def run_server(self, bind_address, port, block=True):
        self.listening_thread = Thread(target=self._run_server, args=(bind_address, port))
        self.listening_thread.start()
        if block:
            self.block()

    def _run_server(self, bind_address, port):
        with server.serve(self._websocket_handler, bind_address, port) as self.websocket_server: 
            log.info(f'Websocket Listing on {bind_address}:{port}')
            self.websocket_server.serve_forever()

    def _signal_handler(self, signum, frame):
        signame = signal.Signals(signum).name
        log.info(f'Stop listening signal handler called with signal {signame} ({signum})')
        self.stop_listening()

    def stop_listening(self):
        self.websocket_server.shutdown()

    def block(self, signals=None):
        signal_handler_installer = SignalHandlerInstaller(signals)
        log.debug('Installing signal handlers')
        signal_handler_installer.install(self._signal_handler)
        try:
            while True:
                # timeout required for windows or we never get the SIGINT signal
                self.listening_thread.join(timeout=0.5)
                if not self.listening_thread.is_alive():
                    break
        finally:
            log.debug('Removing signal handlers')
            signal_handler_installer.remove()
        self.shutdown_server()

    def shutdown_server(self):
        log.info('Starting Server Shutdown')
        self.stop_listening()
        self.listening_thread.join()

        self.time_to_stop = True
        for channel in self.connected_channels.values():
            assert isinstance(channel, ThreadedMsgChannel)
            channel.shutdown_channel()
        log.info('Server Shutdown Complete')

    def _websocket_handler(self, client_websocket):
        websocket_path = client_websocket.request.path
        requested_engine_sig = websocket_path.strip('/')
        remote_address = 'FIXME'
        log.info(f'Accepted connection request from {remote_address} with path {websocket_path}')
        transport = ThreadedWebsocketTransport(client_websocket)
        channel = ThreadedMsgChannel(transport, initiator=False, engine=self.engine, dispatcher=self.dispatcher,
                                       func_registers=self.func_registers)
        self.connected_channels[remote_address] = channel
        channel.start_channel()
        channel.wait_for_remote_close()




import logging
import signal

if 'threaded' in __name__: #= remove
    from threading import Thread #= threaded <
    from websockets.sync import client  #= threaded <
    from websockets.sync import server  #= threaded <
else: #= remove
    import asyncio #= async <
    from websockets.asyncio import client  #= async <
    from websockets.asyncio import server  #= async <

import websockets.exceptions

from ...engine import ProtocolEngine
from ...lib import SignalHandlerInstaller
from ..._template_ import _Template_MsgChannel
from ..transports import _Template_TransportBase
from ...bases import DEFAULT_SERIALISER_SIG, DEFAULT_ENCODER_TAG

log = logging.getLogger(__name__)


class _Template_WebsocketTransport(_Template_TransportBase):
    def __init__(self, websocket, max_msg_size=10*1024*1024, max_bc_length=10, incoming_msg_queue_size=10,
                 outgoing_msg_queue_size=10, socket_buf_size=8192):
        super().__init__(max_msg_size, max_bc_length, incoming_msg_queue_size, outgoing_msg_queue_size, socket_buf_size)
        self.websocket = websocket

    async def _read_data(self, size):
        try:
            return await self.websocket.recv()
        except websockets.exceptions.ConnectionClosed:
            return ''

    async def _write_data(self, data):
        log.debug(f'Sending data: {data}')
        await self.websocket.send(data)

    async def close(self):
        await self.websocket.close()

    
class _Template_WebsocketConnector:
    def __init__(self, engine, dispatcher, default_serialiser_sig: str=DEFAULT_SERIALISER_SIG,
                 default_encoder_tag: str=DEFAULT_ENCODER_TAG, func_registers=None, handshake_cls=None):
        self.engine = engine
        self.dispatcher = dispatcher
        self.default_serialiser_sig = default_serialiser_sig
        self.default_encoder_tag = default_encoder_tag
        self.func_registers = func_registers
        self.initiator = True

    async def connect(self, ws_uri):
        websocket = await client.connect(ws_uri)
        transport = _Template_WebsocketTransport(websocket)  # FIX_ME: Allow options to be set / passed in??
        return _Template_MsgChannel(transport, default_serialiser_sig=self.default_serialiser_sig,
                                    default_encoder_tag=self.default_encoder_tag, initiator=self.initiator,
                                    engine=self.engine, dispatcher=self.dispatcher, func_registers=self.func_registers)


class _Template_WebsocketListener:
    def __init__(self, engine: ProtocolEngine, dispatcher, default_serialiser_sig: str=DEFAULT_SERIALISER_SIG,
                 default_encoder_tag: str=DEFAULT_ENCODER_TAG, func_registers=None):
        self.engine = engine
        self.dispatcher = dispatcher
        self.default_serialiser_sig = default_serialiser_sig
        self.default_encoder_tag = default_encoder_tag
        self.func_registers = func_registers
        self.initiator = False
        self.websocket_server = None
        self.listening_thread = None  #= threaded
        self.listening_task = None  #= async
        self.connected_channels = dict()  # remote_address -> channel
        self.time_to_stop = False

    async def run_server(self, bind_address, port, block=True):
        #= threaded start
        self.listening_thread = Thread(target=self._run_server, args=(bind_address, port))
        self.listening_thread.start()
        #= threaded end
        #= async start
        self.listening_task = asyncio.create_task(self._run_server(bind_address, port))
        #= async end
        if block:
            await self.block()

    async def _run_server(self, bind_address, port):
        async with server.serve(self._websocket_handler, bind_address, port) as self.websocket_server: 
            log.info(f'Websocket Listing on {bind_address}:{port}')
            #= threaded start
            # await is removed during generation from template  #= remove
            await self.websocket_server.serve_forever()
            #= threaded end
            #= async start
            try:
                await self.websocket_server.serve_forever()
            except asyncio.CancelledError:
                pass  # this seems to be raised on server close. Don't re-raise it so we can shut down cleanly.
            #= async end

    def _signal_handler(self, signum, frame):
        signame = signal.Signals(signum).name
        log.info(f'Stop listening signal handler called with signal {signame} ({signum})')
        self.stop_listening()

    def stop_listening(self):
        self.websocket_server.shutdown()  #= threaded
        self.websocket_server.close(close_connections=False)  #= async

    async def block(self, signals=None):
        signal_handler_installer = SignalHandlerInstaller(signals)
        log.debug('Installing signal handlers')
        signal_handler_installer.install(self._signal_handler)
        try:
            #= threaded start
            while True:
                # timeout required for windows or we never get the SIGINT signal
                self.listening_thread.join(timeout=0.5)
                if not self.listening_thread.is_alive():
                    break
            #= threaded end
            #= async start
            await self.listening_task
            #= async end
        finally:
            log.debug('Removing signal handlers')
            signal_handler_installer.remove()
        await self.shutdown_server()

    async def shutdown_server(self):
        log.info('Starting Server Shutdown')
        self.stop_listening()
        self.listening_thread.join() #= threaded
        await self.listening_task #= async

        self.time_to_stop = True
        for channel in self.connected_channels.values():
            assert isinstance(channel, _Template_MsgChannel)
            await channel.shutdown_channel()
        log.info('Server Shutdown Complete')

    async def _websocket_handler(self, client_websocket):
        websocket_path = client_websocket.request.path
        requested_engine_sig = websocket_path.strip('/')
        remote_address = 'FIXME'
        log.info(f'Accepted connection request from {remote_address} with path {websocket_path}')
        transport = _Template_WebsocketTransport(client_websocket)
        channel = _Template_MsgChannel(transport, initiator=False, engine=self.engine, dispatcher=self.dispatcher,
                                       func_registers=self.func_registers)
        self.connected_channels[remote_address] = channel
        await channel.start_channel()
        await channel.wait_for_remote_close()



# -----------------------------------------------------------------------
# WARNING: This file is auto-generated and should not be edited manually!
#
# It is generated using 'generate_from_template_code.py' using the
# _template_ directory as the source.
# -----------------------------------------------------------------------

from __future__ import annotations

from abc import ABC, abstractmethod
from collections import defaultdict
import inspect
import logging
import sys
from typing import Callable

from ..bases import MsgChannelBase, DEFAULT_SERIALISER_SIG, DEFAULT_ENCODER_TAG

from ..registers import FuncInfo
from ..lib import portable_format_exc, IN_MICROPYTHON, ON_WEBASSEMBLY

import asyncio
from asyncio import Lock, Event
try:
    from asyncio import Queue, QueueEmpty
except (AttributeError, ImportError):
    from uasync.queues import Queue, QueueEmpty # type: ignore

# FIXME: add back ThreadPoolExecutor for non-micropython
# also add @non_blocking decorator
#from concurrent.futures import ThreadPoolExecutor


from binarychain import BinaryChain, ChainReader


from ..engine import ProtocolEngine

from ..messages import (FinalType, Acknowledge, ExceptionResponse, Notification, Request, Response,
                        IterableCallbackInfo, MessageSentEvent, ExceptionCategory, ExceptionInfo, CallBase,
                        RequestCallbackInfo, RequestType, ResponseType, MessageBase)

from ..exceptions import CallerException, CalleeException


log = logging.getLogger(__name__)


# noinspection PyUnreachableCode
def get_thread_or_task_name():
    name_getter = getattr(asyncio.current_task(), 'get_name', None)
    if name_getter:
        return name_getter()
    else:
        return 'dummy'


def is_awaitable(may_be_awaitable):
    if sys.implementation.name == 'micropython':
        # FIXME: review inspect once next release of micropython done
        return type(may_be_awaitable).__name__ == 'generator'
    else:
        return inspect.isawaitable(may_be_awaitable)


async def unwrap_result(raw_result):
    # FIXME: We may not need this once we have the @nonblocking decorator
    unwrapped_result = raw_result
    while True:
        if is_awaitable(unwrapped_result):
            unwrapped_result = await unwrapped_result
        else:
            break
    return unwrapped_result


class AsyncTransportBase(ABC):
    def __init__(self, max_msg_size, max_bc_length, incoming_msg_queue_size, outgoing_msg_queue_size, read_buf_size=8192):
        self.max_msg_size = max_msg_size
        self.max_bc_length = max_bc_length
        self.incoming_queue = Queue(maxsize=incoming_msg_queue_size)  # incoming queue of (BinaryChain, complete, remote_closed)
        self.outgoing_queue = Queue(maxsize=outgoing_msg_queue_size)  # outgoing queue of (BinaryChain, outgoing_queue)
        self.read_buf_size = read_buf_size
        self.remote_closed = False
        self.chain_reader = ChainReader(max_part_size=self.max_msg_size, max_chain_size=self.max_msg_size, max_chain_length=self.max_bc_length)
        self.started = False
        self.reader_task = asyncio.create_task(self._reader())
        self.writer_task = asyncio.create_task(self._writer())

    @abstractmethod
    async def _read_data(self, size):
        raise NotImplementedError()

    @abstractmethod
    async def _write_data(self, data):
        raise NotImplementedError()
    
    async def start(self):
        if not self.started:
            self.started = True

    async def _reader(self):
        while True:
            try:
                data = await self._read_data(self.read_buf_size)
                if isinstance(data, (bytearray, bytes)):
                    for incoming_chain in self.chain_reader.get_binary_chains(data):
                        queue_data = (incoming_chain, self.chain_reader.complete(), self.remote_closed)
                        log.debug(f'Put Incoming queue_data: {queue_data}')
                        await self.incoming_queue.put(queue_data)
                else:
                    break
            except Exception as e:
                log.error(f'Reader aborting with exception {e!s}')
                break
        self.remote_closed = True
        try:
            # put sentinel on incoming queue
            queue_data = (None, self.chain_reader.complete(), self.remote_closed)
            log.debug(f'Put Incoming queue_data: (Sentinel) {queue_data}')
            await self.incoming_queue.put(queue_data)
        except Exception as e:
            log.error(f'Error putting sentinel on incoming queue - {e!s}')

    async def _writer(self):
        try:
            while True:
                time_to_stop = await self._writer_send_one()
                if time_to_stop:
                    break
        except Exception as e:
            log.error(f'Unexpected error in _writer - {e!s}')

    async def _writer_send_one(self):
        try:
            queue_data = await self.outgoing_queue.get()
            if queue_data is None:
                return True
            
            chain, notifier_queue = queue_data
            assert isinstance(chain, BinaryChain)
            assert isinstance(notifier_queue, Queue)
            e = None  # return None if not exception
            try:
                data = chain.serialise()
                await self._write_data(data)
            except Exception as exc:
                log.error(f'Exception on sending chain: {chain}')
                log.error(portable_format_exc(exc))
                e = exc
            await notifier_queue.put(e)
            return False
        except Exception as e:
            log.error(f'Writer aborting with exception: {e!s}')
        return True

    # FIXME: Do we even need this now - perhaps just read directly from the queue?
    async def get_next_binary_chain(self):
        log.debug('Starting in get_binary_chains')
        binary_chain, complete, remote_closed = await self.incoming_queue.get()
        return (binary_chain, complete, remote_closed)

    async def send_binary_chain(self, binary_chain):
        log.debug(f'Adding binary chain to outgoing queue: {id(binary_chain)}: {repr(binary_chain)}')
        notifier_queue = Queue()
        queue_item = (binary_chain, notifier_queue)
        await self.outgoing_queue.put(queue_item)
        e = await notifier_queue.get()
        log.debug(f'Binary Chain {id(binary_chain)} sent.')
        if e:
            raise e
        
    async def shutdown(self):
        # signal outgoing queue to stop
        await self.outgoing_queue.put(None)
        queue_data = (None, None, None)
        await self.incoming_queue.put(queue_data)  # end incoming queue
        await self.close()
        try:
            await self.reader_task
        except Exception as e:
            log.error(f'reader_task error:\n{portable_format_exc(e)}')
        try:
            await self.writer_task
        except Exception as e:
            log.error(f'writer_task error:\n{portable_format_exc(e)}')
        
    @abstractmethod
    async def close(self):
        raise NotImplementedError()


class AsyncDispatcherBase(ABC):
    @abstractmethod
    async def dispatch_incoming(self, msg_channel: AsyncMsgChannel, request: CallBase, func_info: FuncInfo):
        raise NotImplementedError()
    @abstractmethod
    async def shutdown(self, timeout=0):
        raise NotImplementedError()


# FIXME: Turn this into a class
async def async_call_func(msg_channel: AsyncMsgChannel, incoming_msg: Request|Notification, func_info: FuncInfo):
    log.debug(f'In async_call_func with {incoming_msg} and {func_info}')
    params = dict()
    injectors = list()
    if func_info.injectable_params:
        for param_name, injector_cls in func_info.injectable_params.items():
            assert issubclass(injector_cls, AsyncInjectorBase)
            injector = injector_cls(msg_channel, incoming_msg, func_info)
            injectors.append(injector)
            await injector.pre_call_setup()
            params[param_name] = await injector.get_param()

    if func_info.func:
        func = func_info.func
    else:
        log.debug('Calling Iterable __anext__')
        # noinspection PyUnresolvedReferences
        func = func_info.iterable_callback.__anext__

    # normal non-binary request
    if incoming_msg.params:
        for param_name, value in incoming_msg.params.items():
            if param_name in params:
                raise ValueError('duplicate param name')  # FIXME: make a caller error
            params[param_name] = value
    if isinstance(incoming_msg, Request) and incoming_msg.callback_params:
        for (param_name, cb_data) in incoming_msg.callback_params.items():
            if param_name in params:
                raise ValueError('duplicate param name')  # FIXME: make a caller error
            assert isinstance(cb_data, dict)
            proxy_type, cb_param_data = list(cb_data.items())[0]
            callback_request_id = cb_param_data['callback_request_id']
            cb_info = cb_param_data['cb_info']
            if proxy_type == '#cb':
                cb_proxy = AsyncCallbackProxy(param_name, callback_request_id, cb_info)
            elif proxy_type == '#icb':
                cb_proxy = AsyncIterableCallbackProxy(param_name, callback_request_id, cb_info)
            else:
                raise ValueError(f'Invalid callback proxy type {proxy_type}')
            assert isinstance(cb_proxy, CallbackProxyBase)
            cb_proxy.set_channel(msg_channel)
            params[param_name] = cb_proxy
    try:
        raw_result = func(**params)  # FIXME - should put in thread pool if not flagged as non-blocking.
        result = await unwrap_result(raw_result)
    except StopAsyncIteration:
        # pass this up, but don't log as this is normal for stopping multipart requests
        raise
    except Exception as e:
        log.warning('-' * 20)
        log.warning(portable_format_exc(e))
        log.warning('-' * 20)
        raise

    # FIXME: Currently the post_call_cleanup is missed if there is an exception
    for injector in injectors:
        assert isinstance(injector, AsyncInjectorBase)
        await injector.post_call_cleanup()
    return result

async def async_run_request_wait_to_complete(msg_channel: AsyncMsgChannel, request: Request,
                                                  func_info: FuncInfo, completed_callback=None):
    try:
        if func_info.multipart_response:
            async for part_result in await async_call_func(msg_channel, request, func_info):
                outgoing_msg = Response(request_id=request.message_id, result=part_result,
                                        response_type=ResponseType.MULTIPART,
                                        envelope_serialisation_code=request.envelope_serialisation_code,
                                        payload_serialisation_code=request.payload_serialisation_code,
                                        protocol_version=request.protocol_version)
                await msg_channel._send_message(outgoing_msg)
            final_response = Response(request_id=request.message_id, response_type=ResponseType.MULTIPART,
                                      final=FinalType.TERMINATOR,
                                      envelope_serialisation_code=request.envelope_serialisation_code,
                                      payload_serialisation_code=request.payload_serialisation_code,
                                      protocol_version=request.protocol_version)
            await msg_channel._send_message(final_response)
        else:
            result = await async_call_func(msg_channel, request, func_info)
            if request.request_type == RequestType.QUIET:
                # never send the result to quiet requests.
                result = None
            outgoing_msg = Response(request_id=request.message_id, result=result,
                                    envelope_serialisation_code=request.envelope_serialisation_code,
                                    payload_serialisation_code=request.payload_serialisation_code,
                                    protocol_version=request.protocol_version)
            await msg_channel._send_message(outgoing_msg)
    except Exception as e:
        if func_info.iterable_callback and isinstance(e, StopAsyncIteration):
            outgoing_msg = Response(request_id=request.message_id, result=None, final=FinalType.TERMINATOR,
                                    envelope_serialisation_code=request.envelope_serialisation_code,
                                    payload_serialisation_code=request.payload_serialisation_code,
                                    protocol_version=request.protocol_version)
        else:
            log.debug(portable_format_exc(e))
            if isinstance(e, CallerException):
                exc_info = ExceptionInfo(category=ExceptionCategory.CALLER, exc_type=type(e).__name__, msg=str(e))
            else:
                exc_info = ExceptionInfo(category=ExceptionCategory.CALLEE, exc_type=type(e).__name__, msg=str(e))
            outgoing_msg = ExceptionResponse(request_id=request.message_id, exc_info=exc_info,
                                             envelope_serialisation_code=request.envelope_serialisation_code,
                                             protocol_version=request.protocol_version
                                             )
        await msg_channel._send_message(outgoing_msg)

    if completed_callback:
        task = asyncio.current_task()
        try:
            await completed_callback(task, request)
        except Exception as e:
            log.error(f'Error calling completed_callback: {e!s}')


async def async_run_request_not_waiting(msg_channel: AsyncMsgChannel, request: Request|Notification,
                                             func_info: FuncInfo, completed_callback=None):
    try:
        # we send a response (with no data) to silent requests (before calling the function), but nothing for notifications.
        if isinstance(request, Request) and request.request_type == RequestType.SILENT:
            outgoing_msg = Response(request_id=request.message_id)
            await msg_channel._send_message(outgoing_msg)

        await async_call_func(msg_channel, request, func_info)
    except Exception as e:
        # FIXME - make notification and silent errors available to the client software
        log.warning(f'Notification or Silent Request {request} raised error {str(e)}')

    if completed_callback:
        task = asyncio.current_task()
        try:
            await completed_callback(task, request)
        except Exception as e:
            log.error(f'Error calling completed_callback: {e!s}')


async def async_dispatch_request_or_notification(msg_channel, incoming_msg, func_info, completed_callback=None):
    try:
        if isinstance(incoming_msg, Request):
            if incoming_msg.request_type == RequestType.SILENT:
                await async_run_request_not_waiting(msg_channel, incoming_msg, func_info, completed_callback)
            else:
                await async_run_request_wait_to_complete(msg_channel, incoming_msg, func_info, completed_callback)
        elif isinstance(incoming_msg, Notification):
            await async_run_request_not_waiting(msg_channel, incoming_msg, func_info, completed_callback)
        else:
            log.warning(f'Got unhandled message {incoming_msg}')
    except Exception as e:
        # FIXME: What do we do here!
        log.error(f'>>>> Dispatch request error: {str(e)} <<<<')


class AsyncDispatcher(AsyncDispatcherBase):
    def __init__(self, num_threads, queue_size=10):
        if num_threads < 1:
            raise ValueError('num_threads must be at least 1')
        self.incoming_queue = Queue(queue_size)
        self.time_to_stop = False

        self.tasks = set()
        self.completed_tasks_queue = Queue()  # using this since MicroPython does not have add_done_callback on tasks
        #self.executor = ThreadPoolExecutor(max_workers=num_threads)  # FIXME: Not using this yet. For non-async functions exported
        self.task_fetcher_task = asyncio.create_task(self._async_task_fetcher())
        self.completed_task_awaiter_task = asyncio.create_task(self._cleanup_completed_tasks())

    async def _async_task_fetcher(self):
        while not self.time_to_stop:
            try:
                msg_channel, incoming_msg, func_info = await asyncio.wait_for(self.incoming_queue.get(), timeout=1)
            except Exception as e:
                if not isinstance(e, asyncio.TimeoutError):
                    log.error(f'*** Got error {str(e)} in _async_task_fetcher - continuing')
                continue
            try:
                task = asyncio.create_task(async_dispatch_request_or_notification(msg_channel, incoming_msg, func_info,
                                                                                        completed_callback=self._task_done_callback))
                self.tasks.add(task)
            except Exception:
                log.error(f'Error creating task for {incoming_msg}!')

    async def _cleanup_completed_tasks(self):
        while not self.time_to_stop:
            try:
                task = await asyncio.wait_for(self.completed_tasks_queue.get(), timeout=1)
            except Exception as e:
                if not isinstance(e, asyncio.TimeoutError):
                    log.error(f'Got exception {e!s} waiting for completed tasks')
                continue

            try:
                # await the completed task to get any exceptions
                await task
                log.debug(f'Task {task!s} awaited')
            except Exception as e:
                log.warning(f'Exception happened in task {task!s}')
                log.warning(portable_format_exc(e))

            try:
                self.tasks.remove(task)
                log.debug(f'Task {task!s} clean up complete')
            except KeyError:
                log.error(f'ERROR: task {task!s} not found!')

    async def _task_done_callback(self, task, incoming_message):
        log.debug(f'**** Task {task!s} for {incoming_message} is done. ****')
        try:
            await self.completed_tasks_queue.put(task)
        except Exception:
            log.error(f'ERROR: Unable to put task {task} onto completed_tasks_queue.')

    async def dispatch_incoming(self, msg_channel: AsyncMsgChannel, request: Request|Notification,
                                func_info: FuncInfo):
        queue_item = (msg_channel, request, func_info)
        await self.incoming_queue.put(queue_item)

    async def shutdown(self, timeout=10):  # FIXME: What timeout do we really want?
        log.debug('Shutting down dispatcher')
        self.time_to_stop = True
        await self.task_fetcher_task
        await self.completed_task_awaiter_task
        for task in self.tasks:
            try:
                task.cancel()
                await task
            except Exception as e:
                log.error(f'Exception in task {task} - {str(e)}')
        log.debug('Dispatcher shut down')


class AsyncMsgChannel(MsgChannelBase):
    def __init__(self, transport: AsyncTransportBase, initiator: bool, engine: ProtocolEngine,
                 dispatcher: AsyncDispatcherBase, default_serialiser_sig: str=DEFAULT_SERIALISER_SIG,
                 default_encoder_tag: str=DEFAULT_ENCODER_TAG,
                 channel_tag='', func_registers=None, channel_register=None):
        super().__init__(initiator, engine, default_serialiser_sig, default_encoder_tag,
                         channel_tag, func_registers, channel_register)
        self.transport = transport
        self.dispatcher = dispatcher
        self.request = AsyncRequestProxyMaker(self)
        self._message_id_lock = Lock()
        self._msg_reader_task = None
        
    def get_proxy(self):
        return AsyncChannelProxy(self)
    
    async def start_channel(self):
        await self.transport.start()
        self._msg_reader_task = asyncio.create_task(self._incoming_msg_pump())
        # asyncio.ensure_future(self._msg_reader_task)
        self._channel_register.register(self)
        log.debug(f'Channel {self} registered')

    async def wait_for_remote_close(self):
        log.debug(f'Waiting for incoming message pump to finish on {get_thread_or_task_name()}')
        await self._msg_reader_task

    async def _incoming_msg_pump(self):
        try:
            log.debug(f'message pump started on thread {get_thread_or_task_name()}')
            last_complete = True
            while True:
                (bin_chain, complete, remote_closed) = await self.transport.incoming_queue.get()
                if bin_chain is None:
                    # NOTE: complete is None is okay, as that is the shutdown marker
                    if not last_complete:
                        log.warning('Got end of chains but not complete')
                    break

                last_complete = complete
                message, dispatch, incoming_callback = self._parse_and_allocate_bin_chain(bin_chain)
                log.debug(f'** Incoming message: {message}, Dispatch: {dispatch}, Callback: {incoming_callback}')
                if dispatch:
                    await self._dispatch(message)
                if incoming_callback:
                    await incoming_callback(message)
        except Exception as e:
            log.error(f'_incoming_msg_pump stopped with exception {e!s}')
        log.debug(f'message pump stopped on thread {get_thread_or_task_name()}')

    async def _dispatch(self, message: Request|Notification):
        func_info, ack_err_msg = self._get_func_info_and_ack_err_msg(message)
        if ack_err_msg:
            await self._send_message(Acknowledge(request_id=message.message_id))
        if func_info:
            await self.dispatcher.dispatch_incoming(self, message, func_info)

    async def _create_message_id(self):
        async with self._message_id_lock:
            message_id = self._message_id
            self._message_id += 1
            return message_id

    async def _send_message(self, message: MessageBase, message_event_callback = None,
                            add_id_callback: Callable = None):
        if hasattr(message, 'message_id') and message.message_id is not None:
                raise ValueError('message_id must not be pre-set')

        if message.envelope_serialisation_code is None:
            message.envelope_serialisation_code = self.default_serialiser_sig

        if message.protocol_version is None:
            message.protocol_version = self.default_encoder_tag

        add_id, register_event_callback = self._get_add_id_and_reg_cb(message)
        if add_id:
            message_id = await self._create_message_id()
            if add_id_callback:
                add_id_callback(message_id)
        else:
            message_id = None
        if register_event_callback:
            self._reg_callback(message_id, message_event_callback)

        if message_id:
            message.message_id = message_id

        bc = self.engine.message_to_binary_chain(message)
        await self.transport.send_binary_chain(bc)
        if message_event_callback:
            await message_event_callback(MessageSentEvent(request_id=message_id))
        return message_id
    
    async def queue_message(self, message, message_event_callback: Callable, add_id_callback: Callable = None):
        return await self._send_message(message, message_event_callback, add_id_callback)

    async def send_shutdown_pending(self):
        # FIXME
        pass

    async def shutdown_channel(self):
        # FIXME: send ?
        self._channel_register.unregister(self)
        await self.transport.shutdown()
        await self._msg_reader_task
        log.debug(f'Channel {self} unregistered and msg_reader thread cleaned up')


# FIXME: Test timeouts!
class AsyncChannelProxy:
    def __init__(self, channel: AsyncMsgChannel):
        self.channel = channel
        self._callbacks = defaultdict(dict)   # self._callbacks[request_id][param_name] = cb_info

    async def send_request_multipart_result_as_iterator(self, message: Request, timeout=None, msg_sent_callback=None, ack_callback=None):
        response_pump = AsyncResponsePump(self, msg_sent_callback, ack_callback, expected_response_type=ResponseType.MULTIPART)
        await response_pump.send_message(message, timeout)
        return response_pump

    async def send_request(self, message: Request, timeout=None, msg_sent_callback=None, ack_callback=None):
        try:
            response_pump = AsyncResponsePump(self, msg_sent_callback, ack_callback)
            await response_pump.send_message(message, timeout)
            result = await response_pump.__anext__()
            # FIXME: Check complete etc?
            return result
        except Exception as e:
            log.error(f'Exception in send_request. {e!s}')
            raise e

    async def send_request_for_iter(self, message: Request, timeout=None, msg_sent_callback=None, ack_callback=None):
        response_pump = AsyncResponsePump(self, msg_sent_callback, ack_callback, callback_iterable=True)
        await response_pump.send_message(message, timeout)
        result = await response_pump.__anext__()
        # FIXME: Check complete etc?
        return result
    
    async def send_notification(self, message: Notification, timeout=None):
        await self._send_message_wait_for_sent_event(message, timeout)

    async def _send_message_wait_for_sent_event(self, message, timeout=None):
        wait_event = Event()
        async def msg_sent_cb(event):
            wait_event.set()
        await self.channel.queue_message(message, msg_sent_cb)
        try:
            await asyncio.wait_for(wait_event.wait(), timeout=timeout)
        except asyncio.TimeoutError:
            log.error('Timeout error on notification send')  # FIXME: Do we raise an error?


    async def send_request_raw_async(self, message: Request, event_callback):
        reg_id_for_callbacks = None
        if isinstance(message, Request) and message.callback_params:
            # avoid a possible race by registering the req_id for any callbacks prior to sending the message
            def reg_id_for_callbacks(req_id):
                for param_name, cb_info in message.callback_params.items():
                    self._callbacks[req_id][param_name] = cb_info

        req_id = await self.channel.queue_message(message, event_callback, add_id_callback = reg_id_for_callbacks)
        return req_id
    
    async def send_request_wait_via_callbacks(self, message: Request, on_msg_sent=None, on_ack=None, on_result=None):
        async def event_callback(event):
            if isinstance(event, MessageSentEvent):
                if on_msg_sent:
                    await on_msg_sent(event)
            elif isinstance(event, Acknowledge):
                if on_ack:
                    await on_ack(event)
            elif isinstance(event, Response):
                if on_result:
                    await on_result(event)
            else:
                log.error(f'Unhandled event type {type(event)}!')  # FIXME: should double check we have everything covered 
        return await self.send_request_raw_async(message, event_callback)


# FIXME: Do we just always return the result and event.final??? Makes the API more consistent
class AsyncResponsePump:
    def __init__(self, channel_proxy: AsyncChannelProxy, msg_sent_callback=None, ack_callback=None,
                 expected_response_type: ResponseType=ResponseType.NORMAL,
                 callback_iterable=False):
        self.channel_proxy = channel_proxy
        
        self.msg_sent_callback = msg_sent_callback
        self.ack_callback = ack_callback
        self.expected_response_type = expected_response_type
        self.callback_iterable = callback_iterable

        self.return_queue = Queue()
        self.count = 0
        self.timeout = None
        self.sent_notification = False
        self.complete = False

    async def _cb_reader(self, event):
        await self.return_queue.put(event)
        log.debug(f'Put event {event!s} on return queue')

    async def send_message(self, message: Request, timeout=None):
        self.timeout = timeout
        if isinstance(message, Notification):
            self.sent_notification = True
        await self.channel_proxy.send_request_raw_async(message, self._cb_reader)

    async def call_msg_sent_callback(self, event):
        if self.msg_sent_callback:
            # would prefer to use asyncio.isawaitable to detect, but
            # this is micropython compatible.
            may_be_awaitable = self.msg_sent_callback(event)
            try:
                await may_be_awaitable
            except TypeError:
                pass
        else:
            log.debug(f'Got msg sent event')

    async def call_ack_callback(self, event):
        if self.ack_callback:
            may_be_awaitable = self.ack_callback(event)
            try:
                await may_be_awaitable
            except TypeError:
                pass
        else:
            log.info(f'Incoming ack received, but no callback passed in')

    async def process_response(self, event):
        # check we have an expected NORMAL or MULTIPART response
        if event.response_type != self.expected_response_type:
            raise ValueError(f'Expected response type {self.expected_response_type} but got {event.response_type}')
        if event.response_type == ResponseType.NORMAL:
            if self.callback_iterable:
                return event.result, event.final
            else:
                return event.result
        elif event.response_type == ResponseType.MULTIPART:
            if event.final == FinalType.FINAL:
                self.complete = True
                return event.result
            elif event.final == FinalType.TERMINATOR:
                self.complete = True
                raise StopAsyncIteration()
            else:
                return event.result
        raise RuntimeError(f'Invalid response type for event {event!s}')
            
    async def process_incoming_req_or_notification(self, event):
        try:
            func_info = None
            cb_info = self.channel_proxy._callbacks[event.callback_request_id][event.target]
            if isinstance(cb_info, RequestCallbackInfo):
                func_info = FuncInfo(target_name=event.target, func=cb_info.func)
            elif isinstance(cb_info, IterableCallbackInfo):
                func_info = FuncInfo(target_name=event.target, iterable_callback=cb_info.iter)
            else:
                log.error(f'Invalid cb_info type of {type(cb_info)}')
            log.debug(f'Calling dispatch with {func_info}')
            await async_dispatch_request_or_notification(self.channel_proxy.channel, event, func_info)
        except KeyError:
            log.error(f'No callback found for incoming callback request {event}')

    async def process_incoming_exception(self, event):
        exc_info = event.exc_info
        assert isinstance(exc_info, ExceptionInfo)
        e = None
        if exc_info.category == ExceptionCategory.CALLER:
            for cls in CallerException.get_subclasses():
                if exc_info.exc_type == cls.__name__:
                    raise cls(exc_info.msg)
            raise CallerException(f'Type: {exc_info.exc_type}, Msg: {exc_info.msg}')
        elif exc_info.category == ExceptionCategory.CALLEE:
            raise CalleeException(f'Type: {exc_info.exc_type}, Msg: {exc_info.msg}')

    async def get_next_event_with_timeout(self):
        while True:
            try:
                log.debug('Waiting for return_queue...')
                if IN_MICROPYTHON and ON_WEBASSEMBLY:
                    # FIXME: work around a crash in micropython on webassembly (pyscript)
                    # means we don't have timeout support (which is not fully developed yet anyway....)
                    # Can be removed once https://github.com/pyscript/pyscript/issues/2415 is fixed in production
                    result = await self.return_queue.get()
                else:
                    result = await asyncio.wait_for(self.return_queue.get(), timeout=1)
                log.debug(f'Waiting for return_queue. Got result: {result}')
                return result
            except Exception as e:
                if not isinstance(e, asyncio.TimeoutError):
                    log.error(f'Get error {str(e)} waiting for timeout')
                self.count += 1
                if self.timeout and self.count > self.timeout:
                    raise RuntimeError('TIMEOUT')  # FIXME. We should at least clean up incoming registers etc. Do we notify the callee?

    def __aiter__(self):
        return self

    async def __anext__(self):
        try:
            while True:
                log.debug(f'Waiting for event in __anext__')
                try:
                    event = await self.get_next_event_with_timeout()
                except Exception as e:
                    log.error(f'Got exception {str(e)} in __anext__. Continuing...')
                    continue

                log.debug(f'Got event in __anext__: {event}')
                if isinstance(event, MessageSentEvent):
                    await self.call_msg_sent_callback(event)
                    if self.sent_notification:
                        self.complete = True
                        raise StopAsyncIteration
                    continue

                if isinstance(event, Acknowledge):
                    await self.call_ack_callback(event)
                    continue

                if isinstance(event, Response):
                    return await self.process_response(event)

                if isinstance(event, Request) or isinstance(event, Notification):
                    await self.process_incoming_req_or_notification(event)
                    continue

                if isinstance(event, ExceptionResponse):
                    await self.process_incoming_exception(event)

                log.error(f'Unhandled event {event}')
        except Exception as e:
            if not isinstance(e, StopAsyncIteration):
                log.error(f'Got exception {str(e)} in __anext__')
            raise e

class AsyncRequestProxy:
    def __init__(self, msg_channel: AsyncMsgChannel, target: str, namespace: str=None, node: str=None,
                 request_type: RequestType=RequestType.NORMAL, timeout: int=None, msg_sent_callback=None,
                 ack_callback=None, multipart_response=False):
        self._msg_channel = msg_channel
        self._target = target
        self._namespace = namespace
        self._node = node
        self._request_type = request_type
        self._timeout = timeout
        self._msg_sent_callback = msg_sent_callback
        self._ack_callback = ack_callback
        self._multipart_response = multipart_response

    def __call__(self, **kwargs):
        params = dict()
        callback_params = dict()
        for param, value in kwargs.items():
            # not just checking isinstance(value, Iterable) because we don't want lists etc
            if hasattr(value, '__aiter__') and hasattr(value, '__anext__'):
                callback_params[param] = IterableCallbackInfo(iter=value)
            elif callable(value):
                callback_params[param] = RequestCallbackInfo(func=value)
            else:
                params[param] = value
        request = Request(target=self._target, namespace=self._namespace, node=self._node, params=params,
                          callback_params=callback_params, request_type=self._request_type,
                          acknowledge=bool(self._ack_callback))
        channel_proxy = self._msg_channel.get_proxy()
        if self._multipart_response:
            return channel_proxy.send_request_multipart_result_as_iterator(request, timeout=self._timeout,
                                                                           msg_sent_callback=self._msg_sent_callback,
                                                                           ack_callback=self._ack_callback)
        else:
            return channel_proxy.send_request(request, timeout=self._timeout, msg_sent_callback=self._msg_sent_callback,
                                              ack_callback=self._ack_callback)
    

class AsyncRequestProxyMaker:
    def __init__(self, msg_channel: AsyncMsgChannel, namespace: str=None, node: str=None,
                 request_type: RequestType=RequestType.NORMAL, timeout: int=None, msg_sent_callback=None,
                 ack_callback=None, multipart_response=False):
        self._msg_channel = msg_channel
        self._namespace = namespace
        self._node = node
        self._request_type = request_type
        self._timeout = timeout
        self._msg_sent_callback = msg_sent_callback
        self._ack_callback = ack_callback
        self._multipart_response = multipart_response

    def __getattr__(self, target):
        return AsyncRequestProxy(self._msg_channel, target, self._namespace, self._node, self._request_type,
                                      self._timeout, self._msg_sent_callback, self._ack_callback,
                                      self._multipart_response)
    
    def __call__(self, namespace: str=None, node: str=None, request_type: RequestType = RequestType.NORMAL,
                 timeout: int=None, msg_sent_callback=None, ack_callback=None, multipart_response=False):
        return AsyncRequestProxyMaker(self._msg_channel, namespace, node, request_type, timeout, msg_sent_callback,
                                           ack_callback, multipart_response)



class CallbackProxyBase:
    def __init__(self, cb_param_name, callback_request_id, cb_info):
        self._cb_param_name = cb_param_name
        self._callback_request_id = callback_request_id
        self.cb_info = cb_info
        self._channel = None

    def set_channel(self, channel):
        # FIXME: create base MsgChannel in core so we can assert this is a MsgChannel here.
        self._channel = channel


class AsyncCallbackProxy(CallbackProxyBase):
    def __init__(self, cb_param_name, callback_request_id, cb_info):
        super().__init__(cb_param_name, callback_request_id, cb_info)
        self.request_type = None  # can be changed before sending / FIXME: how we do set a default?

    def set_to_notification(self):
        self.request_type = None

    def set_to_request(self, request_type: RequestType):
        self.request_type = request_type

    async def __call__(self, **kwargs):
        if self.request_type:
            outgoing_msg = Request(target=self._cb_param_name, callback_request_id=self._callback_request_id,
                                   params=kwargs, request_type=self.request_type)
        else:
            outgoing_msg = Notification(target=self._cb_param_name, callback_request_id=self._callback_request_id,
                                        params=kwargs)
        log.debug(f'Callback Proxy call: {repr(outgoing_msg)}')
        proxy = self._channel.get_proxy()
        if self.request_type:
            result = await proxy.send_request(outgoing_msg)
            return result
        else:
            await proxy.send_notification(outgoing_msg)
            return None


class AsyncIterableCallbackProxy(CallbackProxyBase):
    def __init__(self, cb_param_name, callback_request_id, cb_info):
        super().__init__(cb_param_name, callback_request_id, cb_info)
        self._final = False

    def __aiter__(self):
        return self

    async def __anext__(self):
        if self._final:
            raise StopAsyncIteration()
        
        outgoing_msg = Request(target=self._cb_param_name, callback_request_id=self._callback_request_id)
        log.debug(f'CallbackIterableProxy call: {repr(outgoing_msg)}')
        proxy = self._channel.get_proxy()
        result, final = await proxy.send_request_for_iter(outgoing_msg)
        if final == FinalType.TERMINATOR:
            raise StopAsyncIteration()
        elif final == FinalType.FINAL:
            self._final = True
        return result



class AsyncInjectorBase(ABC):    
    # FIXME: create base MsgChannel in core so we can assert this is a MsgChannel here.
    def __init__(self, msg_channel, incoming_request: CallBase, func_info: FuncInfo):
        self.msg_channel = msg_channel
        self.incoming_request = incoming_request
        self.func_info = func_info

    @abstractmethod
    async def pre_call_setup(self):
        raise NotImplementedError
    
    @abstractmethod
    async def get_param(self):
        raise NotImplementedError
    
    @abstractmethod
    async def post_call_cleanup(self):
        raise NotImplementedError


class AsyncMsgChannelInjector(AsyncInjectorBase):
    async def pre_call_setup(self):
        pass

    # FIXME: create base MsgChannel in core so we can assert this is a MsgChannel here.
    async def get_param(self):
        return self.msg_channel
        
    async def post_call_cleanup(self):
        pass



from __future__ import annotations

import itertools
import os
import time

import pytest


from listening_server_runner import ListeningServerRunner

import common_data
from common_data import Ports, get_ports, on_win32, win32_connect_list, connect_list

from sparrowrpc.engine import ProtocolEngine
from sparrowrpc.exceptions import CalleeException, CallerException
from sparrowrpc.threaded import ThreadedDispatcher


@pytest.fixture(scope="module")
def ports(start_port):
    return get_ports(start_port)


@pytest.fixture(scope="module", autouse=True)
def run_listening_servers(ports: Ports):
    this_dir = os.path.dirname(os.path.abspath(__file__))
    threaded_server_code = os.path.join(this_dir, 'listening_server_code_threaded.py')
    args = ['--tcp_port', str(ports.threaded_tcp), '--ws_port', str(ports.threaded_ws)]
    if not on_win32:
        args.extend(['--uds_path', ports.threaded_uds])
    threaded_listening_server_runner = ListeningServerRunner(threaded_server_code, args)
    threaded_listening_server_runner.start()

    async_server_code = os.path.join(this_dir, 'listening_server_code_async.py')
    args = ['--tcp_port', str(ports.async_tcp), '--ws_port', str(ports.async_ws)]
    if not on_win32:
        args.extend(['--uds_path', ports.async_uds])
    async_listening_server_runner = ListeningServerRunner(async_server_code, args)
    async_listening_server_runner.start()

    time.sleep(2)  # wait for servers to fully start up    
    yield None
    threaded_listening_server_runner.stop()
    async_listening_server_runner.stop()

connect_to_list = win32_connect_list if on_win32 else connect_list
serialisers = ['j', 'm', 'c']
channel_params = list(itertools.product(connect_to_list, serialisers))


@pytest.fixture(scope="module", params=channel_params)
def channel(request, ports: Ports):
    connect_to, serialiser_sig = request.param
    assert isinstance(connect_to, str)
    engine = ProtocolEngine()
    dispatcher = ThreadedDispatcher(num_threads=5)
    if connect_to.endswith('tcp'):
        from sparrowrpc.threaded.transports import ThreadedTcpConnector
        tcp_connector = ThreadedTcpConnector(engine, dispatcher, default_serialiser_sig=serialiser_sig)
        port = ports._asdict()[connect_to]
        channel = tcp_connector.connect('127.0.0.1', port)
    elif connect_to.endswith('ws'):
        from sparrowrpc.threaded.transports.websockets import ThreadedWebsocketConnector
        ws_connector = ThreadedWebsocketConnector(engine, dispatcher, default_serialiser_sig=serialiser_sig)
        port = ports._asdict()[connect_to]
        channel = ws_connector.connect(f'ws://127.0.0.1:{port}/')
    elif connect_to.endswith('uds'):
        from sparrowrpc.threaded.transports import ThreadedUnixSocketConnector
        uds_connector = ThreadedUnixSocketConnector(engine, dispatcher, default_serialiser_sig=serialiser_sig)
        socket_path = ports._asdict()[connect_to]
        channel = uds_connector.connect(socket_path)
    else:
        raise ValueError(f'Unexpected connect_to value of {connect_to}')
    channel.start_channel()
    yield channel
    channel.shutdown_channel()
    dispatcher.shutdown()


def test_ping(channel):
    result = channel.request(namespace='#sys').ping()
    assert result == 'pong'


def test_hello_world(channel):
    result = channel.request.hello_world()
    assert result == 'Hello world!'


def test_hello_name(channel):
    name = 'Bill'
    result = channel.request.hello_world(name=name)
    assert result == f'Hello {name}!'


def test_multipart_response(channel):
    items = list()
    for item in channel.request(multipart_response=True).multipart_response():
        items.append(item)
    assert items == common_data.MULTIPART_RESPONSE_ITEMS


def test_progress_callback(channel):
    progress_messages = list()
    def progress(message):
        progress_messages.append(message)
    result = channel.request.optional_progress_callback(progress=progress)
    assert result == sum(range(10))
    expected_messages = [f'Added {x} to total' for x in range(10)]
    assert expected_messages == progress_messages


def test_progress_callback_not_used(channel):
    result = channel.request.optional_progress_callback()
    assert result == sum(range(10))


def test_iterable_param(channel):
    data = [1, 10, 30, 3]
    def data_as_generator():
        for item in data:
            yield item
    iterable = data_as_generator()
    result = channel.request.iterable_param(nums=iterable)
    assert result == sum(data)


def test_division(channel):
    result = channel.request.division(a=10, b=2)
    assert result == 5


def test_invalid_param(channel):
    with pytest.raises(CallerException) as execinfo:
        result = channel.request.division(a=10, b=None)


def test_divide_zero(channel):
    with pytest.raises(CallerException) as execinfo:
        result = channel.request.division(a=10, b=0)


def test_callee_error(channel):
    with pytest.raises(CalleeException) as execinfo:
        result = channel.request.division(a=11, b=2)



from .decorators import export, make_export_decorator
from .exceptions import CalleeException, CallerException, ProtocolError, TargetNotFound, InvalidParams
from .registers import FuncInfo, FunctionRegister, default_func_register, MsgChannelRegister, global_channel_register



from __future__ import annotations

import logging

from binarychain import BinaryChain

from .engine import ProtocolEngine
from .serialisers import JsonSerialiser
from .encoders.v050 import V050EncoderDecoder
from .encoders.hs import BasicEncoderDecoder
from .messages import Acknowledge, ExceptionResponse, Notification, Request, Response, ExceptionCategory, ExceptionInfo, CallBase, RequestType, ResponseType
from .registers import FuncInfo, FunctionRegister, default_func_register, global_channel_register


log = logging.getLogger(__name__)


DEFAULT_SERIALISER_SIG = JsonSerialiser.sig
DEFAULT_ENCODER_TAG = V050EncoderDecoder.TAG
BASIC_ENCODER_TAG = BasicEncoderDecoder.TAG


class MsgChannelBase:
    def __init__(self, initiator: bool, engine: ProtocolEngine, default_serialiser_sig: str, default_encoder_tag: str,
                 channel_tag='', func_registers=None, channel_register=None):
        self.initiator = initiator
        self.engine = engine

        if default_serialiser_sig not in self.engine.serialisers.keys():
            raise ValueError(f'Serialiser sig of {default_serialiser_sig} not found in engine')
        self.default_serialiser_sig = default_serialiser_sig

        if default_encoder_tag not in self.engine.encoders.keys():
            raise ValueError(f'Encoder tag of {default_encoder_tag} not found in engine')

        self.default_encoder_tag = default_encoder_tag
        self.tag = channel_tag
        self._channel_register = channel_register if channel_register else global_channel_register

        self.system_register = self.engine.get_system_register()
        self.registers = [self.system_register]

        if func_registers:
            if isinstance(func_registers, list):
                for item in func_registers:
                    if isinstance(item, FunctionRegister):
                        self.registers.append(func_registers)
                    else:
                        raise TypeError()
            elif isinstance(func_registers, FunctionRegister):
                self.registers.append(func_registers)
            else:
                raise TypeError()
        else:
            self.registers.append(default_func_register)

        self._message_id = 1
        self._message_event_callbacks = dict()  # message_id -> callable

    def add_register(self, func_register):
        assert isinstance(func_register, FunctionRegister)
        self.registers.append(func_register)

    def _lookup_func_register(self, target, namespace):
        for func_register in self.registers:
            assert isinstance(func_register, FunctionRegister)
            func_info = func_register.get_method_info(target, namespace)
            if func_info:
                return func_info
        return None

    def _get_add_id_and_reg_cb(self, message):
        add_id = self.engine.always_send_ids
        register_event_callback = False
        if isinstance(message, Request):
            add_id = True
            register_event_callback = True
        if isinstance(message, Acknowledge):
            # can't add ids to outgoing acknowledge
            add_id = False
        if isinstance(message, Notification):
            # can't add ids to outgoing notifications
            add_id = False
        if isinstance(message, Response):
            if message.acknowledge:
                add_id = True
                register_event_callback = True
        return add_id, register_event_callback

    def _reg_callback(self, message_id, message_event_callback):
            if message_id is None:
                raise RuntimeError('invalid state - message id should be set')
            if not message_event_callback:
                raise ValueError('message_event_callback required')
            self._message_event_callbacks[message_id] = message_event_callback

    def _get_func_info_and_ack_err_msg(self, message: CallBase):
        ack_err_msg = None
        func_info = self._lookup_func_register(message.target, message.namespace)
        if func_info:
            assert isinstance(func_info, FuncInfo)
            if isinstance(message, Request) and message.acknowledge:
                if message.request_type == RequestType.SILENT:
                    log.warning(f'Silent Request {message} flagged with Acknowledge')
                else:
                    if message.message_id is None:
                        log.error(f'Incoming request without an id')
                    else:
                        ack_err_msg = Acknowledge(request_id=message.message_id)
        else:
            if isinstance(message, Request):
                exc_info = ExceptionInfo(ExceptionCategory.CALLER, exc_type='TargetNotFound', msg=f'target {message.target} not found')
                ack_err_msg = ExceptionResponse(message.message_id, exc_info=exc_info)
        return func_info, ack_err_msg

    def _parse_and_allocate_bin_chain(self, bin_chain: BinaryChain):
        dispatch = False
        incoming_callback = None
        log.debug(f'Got incoming binary chain: {repr(bin_chain)}')
        message = self.engine.parse_incoming_message(bin_chain)
        log.debug(f'Got incoming message {message}')

        if (isinstance(message, Request) or isinstance(message, Notification)) and not message.callback_request_id:
            dispatch = True
            # record incoming serialisation and encoder and set as new default for outgoing messages
            if message.envelope_serialisation_code != self.default_serialiser_sig:
                self.default_serialiser_sig = message.envelope_serialisation_code
            if message.protocol_version != self.default_encoder_tag:
                self.default_encoder_tag = message.protocol_version
        else:
            request_completed = False
            if (isinstance(message, Request) or isinstance(message, Notification)) and message.callback_request_id:
                request_id = message.callback_request_id
            else:
                assert isinstance(message, (Response, ExceptionResponse, Acknowledge))
                request_id = message.request_id
            if isinstance(message, Response):
                if message.response_type == ResponseType.NORMAL:
                    request_completed = True
                elif message.response_type == ResponseType.MULTIPART:
                    request_completed = message.final
                else:
                    raise RuntimeError(f'Invalid response type of {message.response_type!r}')
            if isinstance(message, ExceptionResponse):
                request_completed = True
            try:
                if request_completed:
                    incoming_callback = self._message_event_callbacks.pop(request_id)
                else:
                    incoming_callback = self._message_event_callbacks[request_id]
            except KeyError:
                log.warning(f'Incoming message {message} with invalid request id')
        return message, dispatch, incoming_callback



import logging
from typing import Dict, Any


from ..messages import Notification, CallBase, Request, Response, ExceptionCategory, ExceptionInfo, RequestType, ResponseType, MessageBase

from .base import EncoderDecoderBase, make_out_except_data
from ..messages import ExceptionResponse
from ..exceptions import ProtocolError

"""
This is the 'Basic' engine - basically a modified version of jsonrpc2.
The majority of the Sparrow RPC features are NOT supported with this engine.

It is expected that this engine will remain constant over time, even as the main protocol evolves and is updated.
This provides a way to perform a handshake between systems prior to knowing what (if any) common core protocol versions
they support.
"""

log = logging.getLogger(__name__)


class BasicEncoderDecoder(EncoderDecoderBase):
    TAG = 'b'
    max_bc_length = 1   # params or result (prefix not counted)

    def __init__(self, protocol_engine):
        EncoderDecoderBase.__init__(self, protocol_engine)

        self._make_map = {Request: self._make_out_request,
                          Notification: self._make_out_notification,
                          Response: self._make_out_resp,
                          ExceptionResponse: self._make_out_except,
                          }
        
        self.channels = list()

    def encode_envelope(self, message: MessageBase):
        try:
            # noinspection PyTypeChecker
            make_method = self._make_map[type(message)]
            return make_method(message)
        except KeyError:
            raise TypeError(f'Invalid message type {type(message)}')
        
    def serialise_payload_data(self, message: CallBase):
        return []

    @staticmethod
    def _check_supported_call(message: CallBase):
        if message.callback_request_id:
            raise ValueError('Callbacks not supported with this engine.')
        if message.payload_serialisation_code and message.payload_serialisation_code != message.envelope_serialisation_code:
            raise ValueError('Different payload serialisation not supported with this engine.')
    
    def _make_out_request(self, message: Request):
        self._check_supported_call(message)
        if message.message_id is None:
            raise ValueError('message id is required on Outgoing Requests')
        if message.request_type != RequestType.NORMAL:
            raise ValueError('Only normal requests supported with this engine')
        if message.acknowledge:
            raise ValueError('Acknowledge not supported with this engine.')
        data = self._make_out_call(message)
        data['id'] = message.message_id
        return data

    @staticmethod
    def _make_out_call(message: CallBase):
        data: Dict[str, Any] = dict(target=message.target)
        if message.namespace:
            data['namespace'] = message.namespace
        if message.node:
            data['node'] = message.node
        if message.params:
            data['params'] = message.params
        return data
        
    def _make_out_notification(self, message: Notification):
        self._check_supported_call(message)
        return self._make_out_call(message)

    @staticmethod
    def _check_supported_response(message: Response):
        if message.acknowledge:
            raise ValueError('Acknowledge not supported with this engine.')
        if message.response_type != ResponseType.NORMAL:
            raise ValueError('Multipart responses not supported with this engine.')

    def _make_out_resp(self, message: Response):
        self._check_supported_response(message)
        return dict(result=message.result, id=message.request_id)

    @staticmethod
    def _make_out_except(message: ExceptionResponse):
        except_data = make_out_except_data(message)
        return dict(error=except_data, id=message.request_id)

    def decode_raw_envelope(self, raw_envelope_data: dict, envelope_serialisation_code: str):
        if not isinstance(raw_envelope_data, dict):
            raise ProtocolError('message must be a map/dict')

        if 'target' in raw_envelope_data:
            # either a Request or Notification
            target = raw_envelope_data['target']
            namespace = raw_envelope_data.get('namespace', '')
            node = raw_envelope_data.get('node')
            params = raw_envelope_data.get('params')
            if 'id' in raw_envelope_data:
                return Request(target=target, namespace=namespace, node=node, params=params, message_id=raw_envelope_data['id'],
                               envelope_serialisation_code=envelope_serialisation_code, protocol_version=self.TAG)
            else:
                return Notification(target=target, namespace=namespace, node=node, params=raw_envelope_data.get('params'), 
                                    envelope_serialisation_code=envelope_serialisation_code, protocol_version=self.TAG)
        if 'result' in raw_envelope_data:
            return Response(request_id=raw_envelope_data['id'], result=raw_envelope_data['result'],
                            envelope_serialisation_code=envelope_serialisation_code,
                            protocol_version=self.TAG)
        if 'error' in raw_envelope_data:
            raw_except_data = raw_envelope_data['error']
            exc_info = ExceptionInfo(ExceptionCategory(raw_except_data['cat']), raw_except_data['type'], raw_except_data['msg'])
            return ExceptionResponse(request_id=raw_envelope_data['id'], exc_info=exc_info)
        raise RuntimeError('invalid message passed in')
    
    def parse_payload_data(self, message: MessageBase, raw_payload_data: list):
        if raw_payload_data:
            raise ProtocolError('payload data not supported by this protocol version')
        return



from abc import ABC, abstractmethod
import json
import sys
from typing import Any


try:
    if sys.implementation.name == 'micropython':
        import umsgpack as msgpack  # micropython  # type: ignore
    else:
        import msgpack
except ImportError:
        msgpack = None

try:
    import cbor2
except ImportError:
    cbor2 = None


class BaseSerialiser(ABC):
    sig = None
    @abstractmethod
    def serialise(self, obj_data: Any) -> bytes:
        raise NotImplementedError()
    @abstractmethod
    def deserialise(self, bin_data: bytes) -> Any:
        raise NotImplementedError()


class JsonSerialiser(BaseSerialiser):
    sig = 'j'
    def serialise(self, obj_data: Any) -> bytes:
        return json.dumps(obj_data).encode()
    def deserialise(self, bin_data: bytes) -> Any:
        return json.loads(bin_data.decode())


if msgpack:
    class MsgpackSerialiser(BaseSerialiser):
        sig = 'm'
        # using dumps and loads for micropython compatibility
        def serialise(self, obj_data: Any) -> bytes:
            return msgpack.dumps(obj_data) # type: ignore
        def deserialise(self, bin_data: bytes) -> Any:
            return msgpack.loads(bin_data)
else:
    MsgpackSerialiser = None


if cbor2:
    class CborSerialiser(BaseSerialiser):
        sig = 'c'
        def serialise(self, obj_data: Any) -> bytes:
            return cbor2.dumps(obj_data)
        def deserialise(self, bin_data: bytes) -> Any:
            return cbor2.loads(bin_data)
else:
    CborSerialiser = None



class CallerException(Exception):
    @classmethod
    def get_subclasses(cls):
        for subclass in cls.__subclasses__():
            yield from subclass.get_subclasses()
            yield subclass


class TargetNotFound(CallerException):
    pass


class InvalidParams(CallerException):
    pass


class CalleeException(Exception):
    pass


class ProtocolError(Exception):
    pass



from __future__ import annotations

from collections import defaultdict
import inspect
import logging
from typing import Iterable, Callable, Optional



__all__ = ['FuncInfo', 'FunctionRegister', 'default_func_register', 'MsgChannelRegister', 'global_channel_register']


log = logging.getLogger(__name__)


class FuncInfo:
    def __init__(self, target_name: str, namespace: str = '', auth_groups: Optional[list[str]]=None, multipart_response: bool=False,
                 func: Optional[Callable]=None, non_blocking: bool=False, iterable_callback: Optional[Iterable]=None,  
                 injectable_params: Optional[dict]=None):
        self.target_name = target_name
        self.namespace = namespace
        self.auth_groups = auth_groups # FIXME: maybe something more general, like tags.
        self.multipart_response = multipart_response
        self.func= func
        self.non_blocking = non_blocking # set to true for non-async functions that are exported but can be used directly in async code as they don't block
        self.iterable_callback = iterable_callback
        self.injectable_params = injectable_params  # param name to callable that returns the injected param.

        if self.func and self.iterable_callback:
            raise ValueError('only one of func or iterable_callback can be set')


class FunctionRegister:
    def __init__(self, namespace: str=''):
        self.namespace = namespace if namespace else ''
        self._register = dict()   # dict[namespace][target_name] -> FuncInfo
        
    def register_func(self, func, target_name=None, namespace=None, auth_groups=None, multipart_response=False, injectable_params=None, non_blocking=False):
        namespace = '' if namespace is None else namespace
        if not isinstance(namespace, str):
            raise TypeError('namespace must be a string')
        if self.namespace:
            if namespace:
                raise ValueError('namespace pre-set for this register')
            namespace = self.namespace
        else:
            if namespace == '#sys':
                raise ValueError(f'{namespace} is reserved and cannot be used in this register')
        if not target_name:
            func_data = dict(inspect.getmembers(func))
            target_name = func_data['__name__']
        if auth_groups is None:
            auth_groups = []
        if namespace not in self._register:
            self._register[namespace] = dict()
        func_info = FuncInfo(target_name=target_name, namespace=namespace, auth_groups=auth_groups, multipart_response=multipart_response, func=func,
                             injectable_params=injectable_params, non_blocking=non_blocking)
        if target_name in self._register[namespace]:
            raise ValueError(f'duplicate registration of {target_name} into "{namespace}" namespace')
        self._register[namespace][target_name] = func_info
        log.debug(f'Registered {func_info}')

    def get_method_info(self, target_name, namespace=None) -> Optional[FuncInfo]:
        try:
            return self._register[namespace][target_name]
        except KeyError:
            return None


default_func_register = FunctionRegister()


class MsgChannelRegister:
    def __init__(self):
        self.channel_register = defaultdict(set)  # tag -> set of MsgChannels

    def register(self, msg_channel: MsgChannelBase):
        self.channel_register[msg_channel.tag].add(msg_channel)

    def unregister(self, msg_channel: MsgChannelBase):
        self.channel_register[msg_channel.tag].remove(msg_channel)

    def get_channels_by_tag(self, tag):
        return frozenset(self.channel_register[tag])


global_channel_register = MsgChannelRegister()


# typing imports at the end
from sparrowrpc.bases import MsgChannelBase



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, Any

from abc import ABC, abstractmethod

if TYPE_CHECKING:
    from ..messages import MessageBase, ExceptionResponse
    from ..engine import ProtocolEngine


class EncoderDecoderBase(ABC):
    TAG = ''

    def __init__(self, protocol_engine: ProtocolEngine):
        self.protocol_engine = protocol_engine

    # @abstractmethod
    # def outgoing_message_to_binary_chain(self, message: MessageBase):
    #     raise NotImplementedError()
    
    @abstractmethod
    def encode_envelope(self, message: MessageBase) -> Dict[str, Any]:
        raise NotImplementedError()
    
    @abstractmethod
    def serialise_payload_data(self, message: MessageBase) -> list:
        raise NotImplementedError()

    @abstractmethod
    def decode_raw_envelope(self, raw_envelope_data: dict, envelope_serialisation_code: str) -> MessageBase:
        raise NotImplementedError()
    
    @abstractmethod
    def parse_payload_data(self, message: MessageBase, raw_payload_data: list):
        raise NotImplementedError()



def make_out_except_data(message: ExceptionResponse):
    exc_info = message.exc_info
    except_data: Dict[str, Any] = {'cat': str(exc_info.category), 'type': exc_info.exc_type}
    if exc_info.msg:
        except_data['msg'] = exc_info.msg
    if exc_info.details:
        except_data['details'] = exc_info.details
    if exc_info.value is not None:
        except_data['value'] = exc_info.value
    return except_data




from __future__ import annotations

import sys

IN_MICROPYTHON = (sys.implementation.name == 'micropython')
ON_WEBASSEMBLY = (sys.platform == 'webassembly')


if IN_MICROPYTHON:
    signal = socket = None
    from uasync.queues import Queue
else:
    import signal
    import socket

import traceback
from typing import Optional, Iterable, Callable


class SignalHandlerInstaller:
    def __init__(self, signals: Optional[Iterable[signal.Signals]] = ()):
        if not signals:
            if sys.platform == 'win32':
                signals = [signal.SIGINT, signal.SIGTERM, signal.SIGBREAK]
            else:
                signals = [signal.SIGINT, signal.SIGTERM]
        self.signals = signals
        self.saved_handlers = dict()  # signal -> handler

    def install(self, handler: Callable):
        for sig in self.signals:
            self.saved_handlers[sig] = signal.getsignal(sig)
            signal.signal(sig, handler)
    
    def remove(self):
        for sig in self.signals:
            signal.signal(sig, self.saved_handlers[sig])



def detect_unix_socket_in_use(socket_path):
    detecting_socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    try:
        detecting_socket.connect(socket_path)
        detecting_socket.close()
        return True
    except Exception:
        return False


def portable_format_exc(e):
    tb = getattr(e, '__traceback__', None)
    return '\n'.join(traceback.format_exception(type(e), e, tb))





# -----------------------------------------------------------------------
# WARNING: This file is auto-generated and should not be edited manually!
#
# It is generated using 'generate_from_template_code.py' using the
# _template_ directory as the source.
# -----------------------------------------------------------------------


import logging
import signal

import asyncio
from websockets.asyncio import client
from websockets.asyncio import server

import websockets.exceptions

from ...engine import ProtocolEngine
from ...lib import SignalHandlerInstaller
from ...asyncio import AsyncMsgChannel
from ..transports import AsyncTransportBase
from ...bases import DEFAULT_SERIALISER_SIG, DEFAULT_ENCODER_TAG

log = logging.getLogger(__name__)


class AsyncWebsocketTransport(AsyncTransportBase):
    def __init__(self, websocket, max_msg_size=10*1024*1024, max_bc_length=10, incoming_msg_queue_size=10,
                 outgoing_msg_queue_size=10, socket_buf_size=8192):
        super().__init__(max_msg_size, max_bc_length, incoming_msg_queue_size, outgoing_msg_queue_size, socket_buf_size)
        self.websocket = websocket

    async def _read_data(self, size):
        try:
            return await self.websocket.recv()
        except websockets.exceptions.ConnectionClosed:
            return ''

    async def _write_data(self, data):
        log.debug(f'Sending data: {data}')
        await self.websocket.send(data)

    async def close(self):
        await self.websocket.close()

    
class AsyncWebsocketConnector:
    def __init__(self, engine, dispatcher, default_serialiser_sig: str=DEFAULT_SERIALISER_SIG,
                 default_encoder_tag: str=DEFAULT_ENCODER_TAG, func_registers=None, handshake_cls=None):
        self.engine = engine
        self.dispatcher = dispatcher
        self.default_serialiser_sig = default_serialiser_sig
        self.default_encoder_tag = default_encoder_tag
        self.func_registers = func_registers
        self.initiator = True

    async def connect(self, ws_uri):
        websocket = await client.connect(ws_uri)
        transport = AsyncWebsocketTransport(websocket)  # FIX_ME: Allow options to be set / passed in??
        return AsyncMsgChannel(transport, default_serialiser_sig=self.default_serialiser_sig,
                                    default_encoder_tag=self.default_encoder_tag, initiator=self.initiator,
                                    engine=self.engine, dispatcher=self.dispatcher, func_registers=self.func_registers)


class AsyncWebsocketListener:
    def __init__(self, engine: ProtocolEngine, dispatcher, default_serialiser_sig: str=DEFAULT_SERIALISER_SIG,
                 default_encoder_tag: str=DEFAULT_ENCODER_TAG, func_registers=None):
        self.engine = engine
        self.dispatcher = dispatcher
        self.default_serialiser_sig = default_serialiser_sig
        self.default_encoder_tag = default_encoder_tag
        self.func_registers = func_registers
        self.initiator = False
        self.websocket_server = None
        self.listening_task = None
        self.connected_channels = dict()  # remote_address -> channel
        self.time_to_stop = False

    async def run_server(self, bind_address, port, block=True):
        self.listening_task = asyncio.create_task(self._run_server(bind_address, port))
        if block:
            await self.block()

    async def _run_server(self, bind_address, port):
        async with server.serve(self._websocket_handler, bind_address, port) as self.websocket_server: 
            log.info(f'Websocket Listing on {bind_address}:{port}')
            try:
                await self.websocket_server.serve_forever()
            except asyncio.CancelledError:
                pass  # this seems to be raised on server close. Don't re-raise it so we can shut down cleanly.

    def _signal_handler(self, signum, frame):
        signame = signal.Signals(signum).name
        log.info(f'Stop listening signal handler called with signal {signame} ({signum})')
        self.stop_listening()

    def stop_listening(self):
        self.websocket_server.close(close_connections=False)

    async def block(self, signals=None):
        signal_handler_installer = SignalHandlerInstaller(signals)
        log.debug('Installing signal handlers')
        signal_handler_installer.install(self._signal_handler)
        try:
            await self.listening_task
        finally:
            log.debug('Removing signal handlers')
            signal_handler_installer.remove()
        await self.shutdown_server()

    async def shutdown_server(self):
        log.info('Starting Server Shutdown')
        self.stop_listening()
        await self.listening_task

        self.time_to_stop = True
        for channel in self.connected_channels.values():
            assert isinstance(channel, AsyncMsgChannel)
            await channel.shutdown_channel()
        log.info('Server Shutdown Complete')

    async def _websocket_handler(self, client_websocket):
        websocket_path = client_websocket.request.path
        requested_engine_sig = websocket_path.strip('/')
        remote_address = 'FIXME'
        log.info(f'Accepted connection request from {remote_address} with path {websocket_path}')
        transport = AsyncWebsocketTransport(client_websocket)
        channel = AsyncMsgChannel(transport, initiator=False, engine=self.engine, dispatcher=self.dispatcher,
                                       func_registers=self.func_registers)
        self.connected_channels[remote_address] = channel
        await channel.start_channel()
        await channel.wait_for_remote_close()




import logging
import sys

from pyscript import WebSocket

import asyncio
try:
    from asyncio import Queue, QueueEmpty
except (AttributeError, ImportError):
    from uasync.queues import Queue, QueueEmpty # type: ignore

from .asyncio import AsyncMsgChannel
from .asyncio.transports import AsyncTransportBase
from .lib import portable_format_exc

log = logging.getLogger(__name__)


class PyscriptWebsocketTransport(AsyncTransportBase):
    def __init__(self, max_msg_size=10*1024*1024, max_bc_length=10, incoming_msg_queue_size=10, outgoing_msg_queue_size=10, socket_buf_size=8192):
        AsyncTransportBase.__init__(self, max_msg_size, max_bc_length, incoming_msg_queue_size, outgoing_msg_queue_size, socket_buf_size)
        self.ws_url = None
        self.js_ws = None
        self.raw_incoming = Queue()   # queue of raw incoming messages, terminated by None (signals close)
        self.open_complete = Queue()  # None if open successful, or an exception otherwise

    async def connect(self, ws_url):
        self.ws_url = ws_url
        if sys.implementation.name == 'micropython':
            onmessage = self.mpy_js_onmessage
            log.info('In micropython')
        else:
            onmessage = self.js_onmessage
        self.js_ws = WebSocket(url=ws_url, onopen=self.js_onopen, onmessage=onmessage, onclose=self.js_onclose, onerror=self.js_onerror)
        open_result = await self.open_complete.get()
        if open_result:  # an error
            log.error(f'Got an open_result error of {open_result}')
            raise open_result
        else:
            log.debug(f'connect to {self.ws_url} successful')

    async def js_onopen(self, event):
        log.debug(f'Got onopen event: {event!r}')
        await self.open_complete.put(None)

    async def js_onerror(self, event):
        log.debug(f'Got Javascript WS error: {event}')
        # FIXME: Check what state we are in
        # FIXME: better error
        e = RuntimeError('Error connecting')
        await self.open_complete.put(e)

    def mpy_js_onmessage(self, event):
        log.info(f'mpy got incoming data: {event.data!s}')
        loop = asyncio.get_event_loop()
        loop.create_task(self.js_onmessage(event))

    async def js_onmessage(self, event):
        log.info(f'Got incoming data: {event.data!s}')
        try:
            js_data = event.data
            if isinstance(js_data, str):
                log.error(f'Got non-binary data message - it is being dropped!')
            else:
                if sys.implementation.name == 'micropython':
                    incoming_data = bytes.fromhex(js_data.hex())
                    log.debug(f'Got incoming_data: {incoming_data}')
                else:
                    incoming_data = js_data.tobytes()
                await self.raw_incoming.put(incoming_data)
        except Exception as e:
            log.error(f'Error in js_onmessage: {e!s}')

    async def js_onclose(self, event):
        log.debug(f'Got onclose event: {event!r}')
        await self.raw_incoming.put(None)

    async def _read_data(self, size):
        data = await self.raw_incoming.get()
        if data is None:
            return ''
        if not data:
            log.warning(f'Got empty incoming message - it is being dropped')
        log.debug(f'Got incoming data: {data}')
        return data

    async def _write_data(self, data):
        log.info(f'Sending data: {data}')
        try:
            self.js_ws.send(data)
            log.info(f'Send complete: {data}')
        except Exception as e:
            log.error(f'_write_data error: {portable_format_exc(e)}')

    async def close(self):
        await self.js_ws.close()

    
class PyscriptWebsocketConnector:
    def __init__(self, engine, dispatcher, func_registers=None):
        self.engine = engine
        self.dispatcher = dispatcher
        self.func_registers = func_registers
        self.initiator = True

    async def connect(self, ws_url):
        transport = PyscriptWebsocketTransport()  # FIX_ME: Allow options to be set / passed in??
        await transport.connect(ws_url)
        return AsyncMsgChannel(transport, initiator=self.initiator, engine=self.engine, dispatcher=self.dispatcher, func_registers=self.func_registers)
